/*
 * Copyright  2004 David Reveman, Peter Nilsson
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the names of
 * David Reveman and Peter Nilsson not be used in advertising or
 * publicity pertaining to distribution of the software without
 * specific, written prior permission. David Reveman and Peter Nilsson
 * makes no representations about the suitability of this software for
 * any purpose. It is provided "as is" without express or implied warranty.
 *
 * DAVID REVEMAN AND PETER NILSSON DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL DAVID REVEMAN AND
 * PETER NILSSON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: David Reveman <c99drn@cs.umu.se>
 *          Peter Nilsson <c99pnn@cs.umu.se>
 */

#ifdef HAVE_CONFIG_H
#  include "../config.h"
#endif

#include "glcint.h"

#include <math.h>

#define TRANSFORM(surface) \
  (surface->transform)

#define REPEAT(surface) \
  (surface->repeat)

#define PROG_OR_REPEAT(surface) \
  (surface->repeat || surface->programmatic)

#define GLREPEAT(surface, texture) \
  (surface->repeat && texture->repeatable)

#define MANUALREPEAT(surface, texture) \
  (surface->repeat && (!texture->repeatable))

#define ROTATE(surface) \
  (surface->transform && \
   ((surface->transform->m[0][1] != 0.0) || \
   (surface->transform->m[1][0] != 0.0)))

/* This version of composite uses fragment programs for direct
   Porter-Duff compositing. It cannot handle rotating transformations
   and only one of source and mask can be a transformed the other
   surface must be repeating. Function will fall back to regular composite
   function if this isn't the case.

   TODO: Add manual repeat support.
*/
static glc_bool_t
_glc_composite_direct (glc_operator_t op,
                       glc_surface_t *src,
                       glc_surface_t *mask,
                       glc_surface_t *dst,
                       int x_src,
                       int y_src,
                       int x_mask,
                       int y_mask,
                       int x_dst,
                       int y_dst,
                       int width,
                       int height)
{
  glc_texture_t *src_texture;
  glc_texture_t *mask_texture;
  glc_sub_pixel_region_box_t src_region, mask_region, dst_region;
  glc_region_box_t dirty_region;
  glc_point_t src_tl, src_br, mask_tl, mask_br, translate_src,
    translate_mask;
  double src_width, src_height, mask_width, mask_height;
  glc_program_type_t type;

  type = glc_program_type (dst, src, mask);
  
  if (type == GLC_PROGRAM_TYPE_NOT_SUPPORTED)
    return 0;
  
  /* We cannot continue if we have a rotating transformation or
     if both surfaces have transformations or the surface not being
     transformed isn't repeating. */
  if (TRANSFORM (src) || TRANSFORM (mask)) {
    if (ROTATE (src) || ROTATE (mask))
      return 0;
    
    if ((TRANSFORM (src) && (TRANSFORM (mask) || (!PROG_OR_REPEAT (mask)))) ||
        (TRANSFORM (mask) && (TRANSFORM (src) || (!PROG_OR_REPEAT (src)))))
      return 0;
  }

  src_texture = glc_surface_get_texture (src);
  mask_texture = glc_surface_get_texture (mask);

  if (MANUALREPEAT (src, src_texture) || MANUALREPEAT (mask, mask_texture))
    return 0;

  if (!glc_surface_push_current (dst, GLC_CN_SURFACE_DRAWABLE_CURRENT)) {
    glc_surface_pop_current (dst);
    return 0;
  }

  glDisable (GL_SCISSOR_TEST);

  glc_surface_enable_program (type, dst, src, mask, src_texture, mask_texture);

  glc_set_operator (op);
  
  dst->proc_address->gl_active_texture_arb (GL_TEXTURE0_ARB);
  glc_texture_bind (src_texture);
  
  glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

  if (TRANSFORM (src))
    glc_texture_ensure_filter (src_texture, src->filter);
  else
    glc_texture_ensure_filter (src_texture, GLC_FILTER_NEAREST);
    
  glc_texture_ensure_repeat (src_texture, src->repeat);

  dst->proc_address->gl_active_texture_arb (GL_TEXTURE1_ARB);
  glc_texture_bind (mask_texture);
  
  glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
  
  if (TRANSFORM (mask))
    glc_texture_ensure_filter (mask_texture, mask->filter);
  else
    glc_texture_ensure_filter (mask_texture, GLC_FILTER_NEAREST);

  
  /* calculate source area */
  src_region.x1 = src_region.y1 = 0.0;
  src_region.x2 = src->width;
  src_region.y2 = src->height;
  
  if (TRANSFORM (src))
    glc_matrix_transform_sub_pixel_region (src->transform, &src_region);

  src_width = src_region.x2 - src_region.x1;
  src_height = src_region.y2 - src_region.y1;

  translate_src.x = (src_region.x1 < 0.0)? src_region.x1: 0.0;
  translate_src.y = (src_region.y1 < 0.0)? src_region.y1: 0.0;

  src_region.x1 += x_dst;
  src_region.y1 += y_dst;
  src_region.x2 += (x_dst - x_src);
  src_region.y2 += (y_dst - y_src);
  

  /* calculate mask area */
  mask_region.x1 = mask_region.y1 = 0.0;
  mask_region.x2 = mask->width;
  mask_region.y2 = mask->height;

  if (TRANSFORM (mask))
    glc_matrix_transform_sub_pixel_region (mask->transform, &mask_region);

  mask_width = mask_region.x2 - mask_region.x1;
  mask_height = mask_region.y2 - mask_region.y1;

  translate_mask.x = (mask_region.x1 < 0.0)? mask_region.x1: 0.0;
  translate_mask.y = (mask_region.y1 < 0.0)? mask_region.y1: 0.0;
  mask_region.x1 += x_dst;
  mask_region.y1 += y_dst;
  mask_region.x2 += (x_dst - x_mask);
  mask_region.y2 += (y_dst - y_mask);


  /* calculate destination area */
  dst_region.x1 = x_dst;
  dst_region.y1 = y_dst;
  dst_region.x2 = dst_region.x1 + width;
  dst_region.y2 = dst_region.y1 + height;
  
  if (!REPEAT (src))
    glc_intersect_sub_pixel_region (&dst_region, &src_region, &dst_region);

  if (!REPEAT (mask))
    glc_intersect_sub_pixel_region (&dst_region, &mask_region, &dst_region);
  

  /* re-calculate source area */
  if (REPEAT (src)) {
    src_region.y2 = src->height -
      (((y_src % src->height) + (int) (dst_region.y2 - dst_region.y1)) %
       src->height);
    src_region.y1 = (dst_region.y2 - dst_region.y1) + src_region.y2;    
    src_region.x1 = x_src % src->width;
    src_region.x2 = (dst_region.x2 - dst_region.x1) + src_region.x1;
  } else {
    glc_intersect_sub_pixel_region (&src_region, &dst_region, &src_region);

    src_region.x2 = x_src + (src_region.x2 - src_region.x1) - translate_src.x;
    src_region.y2 = y_src + (src_region.y2 - src_region.y1) - translate_src.y;
    src_region.x1 = x_src - translate_src.x;
    src_region.y1 = y_src - translate_src.y;
  }

  /* re-calculate mask area */
  if (REPEAT (mask)) {
    mask_region.y2 = mask->height -
      (((y_mask % mask->height) + (int) (dst_region.y2 - dst_region.y1)) %
       mask->height);
    mask_region.y1 = (dst_region.y2 - dst_region.y1) + mask_region.y2;    
    mask_region.x1 = x_mask % mask->width;
    mask_region.x2 = (dst_region.x2 - dst_region.x1) + mask_region.x1;
  } else {
    glc_intersect_sub_pixel_region (&mask_region, &dst_region, &mask_region);
    
    mask_region.x2 = x_mask + (mask_region.x2 - mask_region.x1) -
      translate_mask.x;
    mask_region.y2 = y_mask + (mask_region.y2 - mask_region.y1) -
      translate_mask.y;
    mask_region.x1 = x_mask - translate_mask.x;
    mask_region.y1 = y_mask - translate_mask.y;
  }

  /* normalize texture coordinates */
  src_tl.x = (src_region.x1 / src_width) * src_texture->texcoord_width;
  src_tl.y = (src_region.y1 / src_height) * src_texture->texcoord_height;

  src_br.x = (src_region.x2 / src_width) * src_texture->texcoord_width;
  src_br.y = (src_region.y2 / src_height) * src_texture->texcoord_height;

  mask_tl.x = (mask_region.x1 / mask_width) * mask_texture->texcoord_width;
  mask_tl.y = (mask_region.y1 / mask_height) * mask_texture->texcoord_height;

  mask_br.x = (mask_region.x2 / mask_width) * mask_texture->texcoord_width;
  mask_br.y = (mask_region.y2 / mask_height) * mask_texture->texcoord_height;

  if (!REPEAT(src)) {
    src_tl.y = src_texture->texcoord_height - src_tl.y;
    src_br.y = src_texture->texcoord_height - src_br.y;
  }
    
  if (!REPEAT(mask)) {
    mask_tl.y = mask_texture->texcoord_height - mask_tl.y;
    mask_br.y = mask_texture->texcoord_height - mask_br.y;
  }
  
  glBegin (GL_QUADS);

  dst->proc_address->gl_multi_tex_coord_2d_arb
    (GL_TEXTURE0_ARB, src_tl.x, src_tl.y);
  dst->proc_address->gl_multi_tex_coord_2d_arb
    (GL_TEXTURE1_ARB, mask_tl.x, mask_tl.y);
  glVertex2d (dst_region.x1, dst_region.y1);
  
  dst->proc_address->gl_multi_tex_coord_2d_arb
    (GL_TEXTURE0_ARB, src_br.x, src_tl.y);
  dst->proc_address->gl_multi_tex_coord_2d_arb
    (GL_TEXTURE1_ARB, mask_br.x, mask_tl.y);
  glVertex2d (dst_region.x2, dst_region.y1);
  
  dst->proc_address->gl_multi_tex_coord_2d_arb
    (GL_TEXTURE0_ARB, src_br.x, src_br.y);
  dst->proc_address->gl_multi_tex_coord_2d_arb
    (GL_TEXTURE1_ARB, mask_br.x, mask_br.y);
  glVertex2d (dst_region.x2, dst_region.y2);

  dst->proc_address->gl_multi_tex_coord_2d_arb
    (GL_TEXTURE0_ARB, src_tl.x, src_br.y);
  dst->proc_address->gl_multi_tex_coord_2d_arb
    (GL_TEXTURE1_ARB, mask_tl.x, mask_br.y);
  glVertex2d (dst_region.x1, dst_region.y2);

  glEnd ();

  glc_texture_unbind (mask_texture);

  dst->proc_address->gl_active_texture_arb (GL_TEXTURE0_ARB);

  glc_texture_unbind (src_texture);

  glc_surface_disable_program (dst);

  dirty_region.x1 = floor (dst_region.x1);
  dirty_region.x1 = floor (dst_region.y1);
  dirty_region.x2 = ceil (dst_region.x2);
  dirty_region.x2 = ceil (dst_region.y2);
  glc_surface_dirty (dst, &dirty_region);
    
  glc_surface_pop_current (dst);

  return 1;
}

static void
glc_mask_bounds (glc_surface_t *src,
                 glc_surface_t *mask,
                 glc_surface_t *dst,
                 int x_src,
                 int y_src,
                 int x_mask,
                 int y_mask,
                 int x_dst,
                 int y_dst,
                 glc_region_box_t *bounds,
                 glc_region_box_t *mbounds)
{
  glc_region_box_t region;
  
  if (bounds->x1 <= 0)
    mbounds->x1 = 0;
  else
    mbounds->x1 = bounds->x1;

  if (bounds->y1 <= 0)
    mbounds->y1 = 0;
  else
    mbounds->y1 = bounds->y1;

  if (bounds->x2 >= dst->width)
    mbounds->x2 = dst->width;
  else
    mbounds->x2 = bounds->x2;

  if (bounds->y2 >= dst->height)
    mbounds->y2 = dst->height;
  else
    mbounds->y2 = bounds->y2;

  if (!REPEAT (src)) {
    region.x1 = x_dst;
    region.y1 = y_dst;
    if (x_src < 0) region.x1 -= x_src;
    if (y_src < 0) region.y1 -= y_src;
    region.x2 = region.x1 + src->width;
    region.y2 = region.y1 + src->height;
    if (x_src > 0) region.x2 -= x_src;
    if (y_src > 0) region.y2 -= y_src;

    if (TRANSFORM (src))
      glc_matrix_transform_region (src->transform, &region);
    
    if (mbounds->x1 < region.x1)
      mbounds->x1 = region.x1;
    
    if (mbounds->y1 < region.y1)
      mbounds->y1 = region.y1;
    
    if (mbounds->x2 > region.x2)
      mbounds->x2 = region.x2;
    
    if (mbounds->y2 > region.y2)
      mbounds->y2 = region.y2;
  }

  if (!REPEAT (mask)) {
    region.x1 = x_dst;
    region.y1 = y_dst;
    if (x_mask < 0) region.x1 -= x_mask;
    if (y_mask < 0) region.y1 -= y_mask;
    region.x2 = region.x1 + mask->width;
    region.y2 = region.y1 + mask->height;
    if (x_mask > 0) region.x2 -= x_mask;
    if (y_mask > 0) region.y2 -= y_mask;
    
    if (TRANSFORM (mask))
      glc_matrix_transform_region (mask->transform, &region);
    
    if (mbounds->x1 < region.x1)
      mbounds->x1 = region.x1;
    
    if (mbounds->y1 < region.y1)
      mbounds->y1 = region.y1;
    
    if (mbounds->x2 > region.x2)
      mbounds->x2 = region.x2;
    
    if (mbounds->y2 > region.y2)
      mbounds->y2 = region.y2;
  }
}

void
glc_composite (glc_operator_t op,
               glc_surface_t *src,
               glc_surface_t *mask,
               glc_surface_t *dst,
               int x_src,
               int y_src,
               int x_mask,
               int y_mask,
               int x_dst,
               int y_dst,
               int width,
               int height)
{
  glc_surface_t *intermediate = NULL, *mask_surface;
  glc_texture_t *texture;
  glc_point_t tl, bl, br, tr;
  glc_region_box_t clip;

  if (src->programmatic)
    glc_programmatic_surface_setup (src,
                                    width + abs (x_src),
                                    height + abs (y_src));

  if (mask && mask->programmatic)
    glc_programmatic_surface_setup (mask,
                                    width + abs (x_mask),
                                    height + abs (y_mask));

  if (mask) {
    glc_region_box_t mask_bounds;
    static glc_color_t clear_color = { 0x0000, 0x0000, 0x0000, 0x0000 };
    glc_bool_t intermediate_translate;

    if ((dst->feature_mask & GLC_FEATURE_ARB_FRAGMENT_PROGRAM_MASK) &&
        _glc_composite_direct (op,
                               src, mask, dst,
                               x_src, y_src,
                               x_mask, y_mask,
                               x_dst, y_dst,
                               width, height))
      return;

    mask_bounds.x1 = x_dst;
    mask_bounds.y1 = y_dst;
    intermediate_translate = 0;
    
    if (!mask->implicit_mask) {
      glc_region_box_t bounds;
      
      bounds.x1 = x_dst;
      bounds.x2 = x_dst + width;
      bounds.y1 = y_dst;
      bounds.y2 = y_dst + height;
      
      glc_mask_bounds (src, mask, dst,
                       x_src, y_src, x_mask, y_mask, x_dst, y_dst,
                       &bounds, &mask_bounds);

      if (mask_bounds.x2 - mask_bounds.x1 <= 0 ||
          mask_bounds.y2 - mask_bounds.y1 <= 0)
        return;

      if ((x_dst != mask_bounds.x1) || (y_dst != mask_bounds.y1))
        intermediate_translate = 1;
      
      mask_surface = intermediate =
        glc_surface_create_similar (dst,
                                    GLC_STANDARD_ARGB32,
                                    mask_bounds.x2 - mask_bounds.x1,
                                    mask_bounds.y2 - mask_bounds.y1);
      
      if (!mask_surface) {
        glc_surface_status_add (dst, GLC_STATUS_NOT_SUPPORTED_MASK);
        return;
      }

      if (TRANSFORM (mask))
        glc_fill_rectangle (GLC_OPERATOR_SRC,
                            mask_surface,
                            &clear_color,
                            0, 0,
                            mask_surface->width,
                            mask_surface->height);

      if ((!REPEAT (mask)) && intermediate_translate) {
        glc_surface_push_transform (mask);
        glc_matrix_translate (mask->transform,
                              x_dst - mask_bounds.x1,
                              y_dst - mask_bounds.y1);
      }
      
      glc_composite (GLC_OPERATOR_SRC,
                     mask, NULL, mask_surface,
                     x_mask, y_mask,
                     0, 0,
                     0, 0,
                     mask_surface->width - (x_dst - mask_bounds.x1),
                     mask_surface->height - (y_dst - mask_bounds.y1));

      if (intermediate_translate)
        glc_surface_pop_transform (mask);
      
    } else
      mask_surface = mask;

    if ((!REPEAT (src)) && intermediate_translate) {
      glc_surface_push_transform (src);
      glc_matrix_translate (src->transform,
                            x_dst - mask_bounds.x1,
                            y_dst - mask_bounds.y1);
    }

    if (TRANSFORM (src))
      mask_surface->clear_exterior = 1;
    
    glc_composite (GLC_OPERATOR_IN,
                   src, NULL, mask_surface,
                   x_src, y_src,
                   0, 0,
                   0, 0,
                   mask_surface->width,
                   mask_surface->height);

    if (intermediate_translate)
      glc_surface_pop_transform (src);

    x_dst = mask_bounds.x1;
    y_dst = mask_bounds.y1;
    width = mask_surface->width;
    height = mask_surface->height;
    src = mask_surface;
    x_src = y_src = 0;
  }

  texture = glc_surface_get_texture (src);

  if (!glc_surface_push_current (dst, GLC_CN_SURFACE_DRAWABLE_CURRENT)) {
    glc_surface_pop_current (dst);
    return;
  }

  glc_texture_bind (texture);

  glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

  clip.x1 = x_dst;
  clip.y1 = y_dst;
  clip.x2 = clip.x1 + width;
  clip.y2 = clip.y1 + height;  
  
  glScissor (clip.x1, dst->height - (clip.y1 + height),
             width, height);
  
  glc_set_operator (op);

  if (src->convolution || src->programmatic)
    glc_surface_enable_program (glc_program_type (dst, src, NULL),
                                dst, src, NULL, texture, NULL);
  
  if ((!TRANSFORM (src)) && GLREPEAT (src, texture)) {
    /* CASE 1: Repeat, no transformation and power of two sized texture,
       GL can do repeat for us. */
    double repeat_factor_x, repeat_factor_y;

    glc_texture_ensure_repeat (texture, 1);
    glc_texture_ensure_filter (texture, GLC_FILTER_NEAREST);
    
    bl.x = tl.x = x_dst;
    tr.y = tl.y = y_dst;
    tr.x = br.x = x_dst + width;
    bl.y = br.y = y_dst + height;
    
    /* Shift coordinates with source offset */
    if (x_src) {
      x_src = (x_src % src->width);
      tl.x -= x_src;
      bl.x -= x_src;
    }
    if (y_src) {
      y_src = (y_src % src->height);
      tl.y -= y_src;
      tr.y -= y_src;
    }

    /* Align with top left edge */
    bl.y += texture->height - (((int) (br.y - tl.y)) % texture->height);
    br.y += texture->height - (((int) (br.y - tl.y)) % texture->height);
      
    repeat_factor_x = (br.x - tl.x) / (double) texture->width;
    repeat_factor_y = (br.y - tl.y) / (double) texture->height;
      
    glBegin (GL_QUADS);
    glTexCoord2d (0.0, repeat_factor_y);
    glVertex2d (tl.x, tl.y);
    glTexCoord2d (repeat_factor_x, repeat_factor_y);
    glVertex2d (tr.x, tr.y);
    glTexCoord2d (repeat_factor_x, 0.0);
    glVertex2d (br.x, br.y);
    glTexCoord2d (0.0, 0.0);
    glVertex2d (bl.x, bl.y);
    glEnd ();
    
  } else {
    /* CASE 2: Either none power of two sized texture or
       transformation is set. */
    double save_tlx, save_trx, save_blx, save_brx;

    glc_texture_ensure_repeat (texture, 0);

    bl.x = tl.x = 0;
    tr.y = tl.y = 0;
    tr.x = br.x = src->width;
    bl.y = br.y = src->height;

    if (TRANSFORM (src)) {
      glc_texture_ensure_filter (texture, src->filter);
      glc_matrix_transform_point (src->transform, &tl);
      glc_matrix_transform_point (src->transform, &bl);
      glc_matrix_transform_point (src->transform, &tr);
      glc_matrix_transform_point (src->transform, &br);
    } else
      glc_texture_ensure_filter (texture, GLC_FILTER_NEAREST);
    
    /* Shift all coordinates with destination offset */
    if (x_dst) {
      tl.x += x_dst;
      bl.x += x_dst;
      tr.x += x_dst;
      br.x += x_dst;
    }
    if (y_dst) {
      tl.y += y_dst;
      bl.y += y_dst;
      tr.y += y_dst;
      br.y += y_dst;
    }

    /* Shift all coordinates with source offset */
    if (x_src) {
      x_src = abs (x_src);
      if (REPEAT (src))
        x_src = (x_src % src->width);
      tl.x -= x_src;
      bl.x -= x_src;
      tr.x -= x_src;
      br.x -= x_src;
    }
    if (y_src) {
      y_src = abs (y_src);
      if (REPEAT (src))
        y_src = (y_src % src->height);
      tl.y -= y_src;
      bl.y -= y_src;
      tr.y -= y_src;
      br.y -= y_src;
    }

    save_tlx = tl.x;
    save_blx = bl.x;
    save_trx = tr.x;
    save_brx = br.x;

    do {
      do {
        /* Clip to original source area if repeat and transform are both
           used. */
        if (TRANSFORM (src) && REPEAT (src)) {
          glc_region_box_t src_clip, intersect_clip;
          
          src_clip.x1 = tl.x;
          src_clip.y1 = tl.y;
          src_clip.x2 = src_clip.x1 + src->width;
          src_clip.y2 = src_clip.y1 + src->height;

          glc_intersect_region (&clip, &src_clip, &intersect_clip);

          glScissor (intersect_clip.x1,
                     dst->height - (intersect_clip.y1 +
                                    (intersect_clip.y2 - intersect_clip.y1)),
                     intersect_clip.x2 - intersect_clip.x1,
                     intersect_clip.y2 - intersect_clip.y1);
        }
        
        glBegin (GL_QUADS);
        glTexCoord2d (0.0, texture->texcoord_height);
        glVertex2d (tl.x, tl.y);
        glTexCoord2d (texture->texcoord_width, texture->texcoord_height);
        glVertex2d (tr.x, tr.y);
        glTexCoord2d (texture->texcoord_width, 0.0);
        glVertex2d (br.x, br.y);
        glTexCoord2d (0.0, 0.0);
        glVertex2d (bl.x, bl.y);
        glEnd ();

        if (REPEAT (src)) {
          bl.x += src->width;
          tl.x += src->width;
          tr.x += src->width;
          br.x += src->width;
        }
        
      } while (REPEAT (src) && (tl.x < (x_dst + width)));

      if (REPEAT (src)) {
        bl.y += src->height;
        tl.y += src->height;
        tr.y += src->height;
        br.y += src->height;

        tl.x = save_tlx;
        bl.x = save_blx;
        tr.x = save_trx;
        br.x = save_brx;
      }
      
    } while (REPEAT (src) && (tl.y < (y_dst + height)));  
  }

  if (src->convolution || src->programmatic)
    glc_surface_disable_program (dst);
  
  glc_texture_unbind (texture);

  /* Clear intermediate mask exterior. This is only done if
     destination surfaces is an intermediate mask surface
     and source surface is transformed and not repeating. */
  if (dst->clear_exterior && (!REPEAT (src))) {
    glc_set_operator (GLC_OPERATOR_SRC);
    glColor4f (0.0, 0.0, 0.0, 0.0);
    
    glBegin (GL_QUADS);

    glVertex2d (tl.x, tl.y);
    glVertex2d (0.0, tl.y);
    glVertex2d (0.0, 0.0);
    glVertex2d (tl.x, 0.0);   

    glVertex2d (tl.x, tl.y);
    glVertex2d (tl.x, 0.0);
    glVertex2d (tr.x, 0.0);
    glVertex2d (tr.x, tr.y);
    
    glVertex2d (tr.x, tr.y);
    glVertex2d (tr.x, 0.0);
    glVertex2d (dst->width, 0.0);
    glVertex2d (dst->width, tr.y);

    glVertex2d (tr.x, tr.y);
    glVertex2d (dst->width, tr.y);
    glVertex2d (dst->width, br.y);
    glVertex2d (br.x, br.y);
    
    glVertex2d (br.x, br.y);
    glVertex2d (dst->width, br.y);
    glVertex2d (dst->width, dst->height);
    glVertex2d (br.x, dst->height);

    glVertex2d (br.x, br.y);
    glVertex2d (br.x, dst->height);
    glVertex2d (bl.x, dst->height);
    glVertex2d (bl.x, bl.y);

    glVertex2d (bl.x, bl.y);
    glVertex2d (bl.x, dst->height);
    glVertex2d (0.0, dst->height);
    glVertex2d (0.0, bl.y);

    glVertex2d (bl.x, bl.y);
    glVertex2d (0.0, bl.y);
    glVertex2d (0.0, tl.y);
    glVertex2d (tl.x, tl.y);

    glEnd ();
  }

  glc_surface_dirty (dst, &clip);

  glc_surface_pop_current (dst);

  if (intermediate)
    glc_surface_destroy (intermediate);
}
