/*
 * Copyright  2004 David Reveman, Peter Nilsson
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the names of
 * David Reveman and Peter Nilsson not be used in advertising or
 * publicity pertaining to distribution of the software without
 * specific, written prior permission. David Reveman and Peter Nilsson
 * makes no representations about the suitability of this software for
 * any purpose. It is provided "as is" without express or implied warranty.
 *
 * DAVID REVEMAN AND PETER NILSSON DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL DAVID REVEMAN AND
 * PETER NILSSON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: David Reveman <c99drn@cs.umu.se>
 *          Peter Nilsson <c99pnn@cs.umu.se>
 */

#ifdef HAVE_CONFIG_H
#  include "../config.h"
#endif

#include "glcint.h"

extern glc_gl_proc_address_list_t _glc_agl_gl_proc_address;

static glc_surface_t *
_glc_agl_surface_create_similar (void *abstract_templ,
                                 glc_format_name_t format_name,
                                 int width,
                                 int height);

static void
_glc_agl_surface_destroy (void *abstract_surface);

static glc_texture_t *
_glc_agl_surface_get_texture (void *abstract_surface);

static void
_glc_agl_surface_realize (void *abstract_surface);

static void
_glc_agl_surface_show (void *abstract_surface);

static glc_bool_t
_glc_agl_surface_push_current (void *abstract_surface,
                               glc_constraint_t constraint)
{
  glc_agl_surface_t *surface = (glc_agl_surface_t *) abstract_surface;
  
  if (constraint == GLC_CN_SURFACE_DRAWABLE_CURRENT &&
      ((!surface->pbuffer) && (!surface->drawable)))
    constraint = GLC_CN_ANY_CONTEXT_CURRENT;
  
  surface = glc_agl_context_push_current (surface, constraint);

  if (surface) {
    glc_surface_setup_environment (&surface->base);
    return 1;
  }

  return 0;
}

static void
_glc_agl_surface_pop_current (void *abstract_surface)
{
  glc_agl_surface_t *surface = (glc_agl_surface_t *) abstract_surface;

  surface = glc_agl_context_pop_current (surface);
  
  if (surface)
    glc_surface_setup_environment (&surface->base);
}

static const struct glc_surface_backend glc_agl_surface_backend = {
  _glc_agl_surface_create_similar,
  _glc_agl_surface_destroy,
  _glc_agl_surface_push_current,
  _glc_agl_surface_pop_current,
  _glc_agl_surface_get_texture,
  _glc_agl_surface_realize,
  _glc_agl_surface_show
};

static void
_glc_agl_surface_ensure_texture (glc_agl_surface_t *surface)
{
  if (!surface->base.dirty)
    return;
    
  if (!surface->pbuffer)
    glc_texture_copy_surface (surface->base.texture, &surface->base,
                              &surface->base.dirty_region);
  
  surface->base.dirty = 0;
}

static glc_texture_t *
_glc_agl_surface_get_texture (void *abstract_surface) {
  glc_agl_surface_t *surface = (glc_agl_surface_t *) abstract_surface;

  if (!surface->base.texture->allocated)
    glc_texture_allocate (surface->base.texture);
  
  _glc_agl_surface_ensure_texture (surface);
  
  return surface->base.texture;
}

static void
_glc_agl_surface_realize_window (WindowRef window,
                                 int *width,
                                 int *height)
{
  Rect window_bounds;

  GetWindowPortBounds (window, &window_bounds);
  
  *width = window_bounds.right - window_bounds.left;
  *height = window_bounds.bottom - window_bounds.top;
}

static void
_glc_agl_set_features (glc_agl_surface_t *surface)
{
  surface->base.feature_mask = surface->thread_info->feature_mask;

  surface->base.feature_mask &= ~GLC_FEATURE_CONVOLUTION_FILTER_MASK;
  surface->base.feature_mask &= ~GLC_FEATURE_MULTISAMPLE_MASK;
  surface->base.feature_mask &= ~GLC_FEATURE_OFFSCREEN_MULTISAMPLE_MASK;

#if defined(GL_ARB_multitexture) && defined(GL_ARB_fragment_program)
  if (surface->thread_info->feature_mask &
      GLC_FEATURE_CONVOLUTION_FILTER_MASK) {
    GLuint texture_indirections;

    surface->base.proc_address->gl_get_program_iv_arb
      (GL_FRAGMENT_PROGRAM_ARB,
       GL_MAX_PROGRAM_TEX_INDIRECTIONS_ARB,
       &texture_indirections);

    /* Convolution filter programs require support for at least six
       texture indirections. */
    if (texture_indirections >= 5)
      surface->base.feature_mask |= GLC_FEATURE_CONVOLUTION_FILTER_MASK;
  }
#endif

  if (surface->base.format->multisample.supported) {
    surface->base.feature_mask |= GLC_FEATURE_MULTISAMPLE_MASK;
    if (surface->pbuffer)
      surface->base.feature_mask |= GLC_FEATURE_OFFSCREEN_MULTISAMPLE_MASK;
  }
}

static glc_surface_t *
_glc_agl_surface_create (glc_agl_thread_info_t *thread_info,
                         glc_format_t *format,
                         int width,
                         int height)
{
  glc_agl_surface_t *surface;
  glc_agl_context_t *context;
  unsigned int texture_format;
  long int texture_target_mask;

  context = glc_agl_context_get (thread_info, format, 1);
  if (!context)
    return NULL;

  surface = (glc_agl_surface_t *) calloc (1, sizeof (glc_agl_surface_t));
  if (surface == NULL)
    return NULL;

  glc_surface_init (&surface->base, &glc_agl_surface_backend);
  
  surface->thread_info = thread_info;
  surface->context = context;

  surface->base.proc_address = &_glc_agl_gl_proc_address;
  surface->base.programs = &thread_info->programs;
  surface->base.feature_mask = 0;
  surface->base.format = format;
  surface->base.width = width;
  surface->base.height = height;

  texture_format = glc_get_gl_format_from_bpp (format->bpp);

  glc_surface_push_current (&surface->base, GLC_CN_ANY_CONTEXT_CURRENT);

  texture_target_mask = thread_info->texture_mask;

  /* Seems to be problem with binding a pbuffer to some power of two sized
     textures. This will try to avoid the problem. */
  if (((width > 1) && (width < 64)) ||
      ((height > 1) && (height < 64)))
    texture_target_mask &= ~GLC_TEXTURE_TARGET_2D_MASK;

  surface->base.texture =
    glc_texture_generate (width, height,
                          texture_format,
                          texture_target_mask);

  if (!surface->base.texture) {
    glc_surface_pop_current (&surface->base);
    glc_surface_destroy (&surface->base);
    return NULL;
  }
  
  if (thread_info->feature_mask & GLC_FEATURE_OFFSCREEN_DRAWING_MASK)
    surface->pbuffer = glc_agl_pbuffer_create (surface->base.texture);
  
  _glc_agl_set_features (surface);

  if (!surface->pbuffer) {
    glc_texture_allocate (surface->base.texture);
  } else {
    glc_surface_push_current (&surface->base, GLC_CN_SURFACE_CONTEXT_CURRENT);
    glc_agl_pbuffer_bind (surface->pbuffer,
                          surface->context->context,
                          surface->base.texture,
                          surface->base.format);
    glc_surface_pop_current (&surface->base);
  }

  glc_surface_pop_current (&surface->base);

  return &surface->base;
}

glc_surface_t *
glc_agl_surface_create (glc_format_t *format,
                        int width,
                        int height)
{
  return _glc_agl_surface_create (glc_agl_thread_info_get (),
                                  format, width, height);
}
slim_hidden_def(glc_agl_surface_create_offscreen);

glc_surface_t *
glc_agl_surface_create_for_window (glc_format_t *format,
                                   WindowRef window)
{
  glc_agl_surface_t *surface;
  glc_agl_context_t *context;
  int width, height;
  glc_agl_thread_info_t *thread_info = glc_agl_thread_info_get ();

  context = glc_agl_context_get (thread_info, format, 0);
  if (!context)
    return NULL;

  _glc_agl_surface_realize_window (window, &width, &height);

  surface = (glc_agl_surface_t *) calloc (1, sizeof (glc_agl_surface_t));
  if (surface == NULL)
    return NULL;

  glc_surface_init (&surface->base, &glc_agl_surface_backend);
  
  surface->thread_info = thread_info;
  surface->context = context;

  surface->base.proc_address = &_glc_agl_gl_proc_address;
  surface->base.programs = &thread_info->programs;
  surface->base.format = format;
  surface->base.width = width;
  surface->base.height = height;

  _glc_agl_set_features (surface);

  surface->window = window;
  surface->drawable = GetWindowPort (window);
  
  return &surface->base;
}
slim_hidden_def(glc_agl_surface_create_for_window);

static glc_surface_t *
_glc_agl_surface_create_similar (void *abstract_templ,
                                 glc_format_name_t format_name,
                                 int width,
                                 int height)
{
  glc_agl_surface_t *templ = (glc_agl_surface_t *) abstract_templ;
  
  if (templ->thread_info->agl_feature_mask & GLC_AGL_FEATURE_PBUFFER_MASK) {
    glc_format_t *format;

    format = glc_format_find_sufficient_standard
      (templ->base.format, 1, GLC_FORMAT_OPTION_OFFSCREEN_MASK, format_name);
    
    if (!format)
      format = glc_format_find_standard (templ->thread_info->formats,
                                         templ->thread_info->n_formats,
                                         GLC_FORMAT_OPTION_OFFSCREEN_MASK,
                                         format_name);

    if (format)
      return _glc_agl_surface_create (templ->thread_info, format,
                                      width, height);
  }

  return NULL;
}

static void
_glc_agl_surface_destroy (void *abstract_surface)
{
  glc_agl_surface_t *surface = (glc_agl_surface_t *) abstract_surface;
  AGLContext context = aglGetCurrentContext ();
  
  if (context == surface->context->context) {
    if (surface->pbuffer) {
      AGLPbuffer pbuffer;
      GLuint unused;
      
      aglGetPBuffer (context, &pbuffer, &unused, &unused, &unused);
      
      if (pbuffer == surface->pbuffer)
        glc_agl_context_make_current (surface);
    } else if (surface->drawable) {
      if (aglGetDrawable (context) == surface->drawable)
        glc_agl_context_make_current (surface);
    }
  }
  
  if (surface->base.texture)
    glc_texture_destroy (surface->base.texture);
  
  if (surface->pbuffer)
    glc_agl_pbuffer_destroy (surface->pbuffer);
  
  glc_surface_deinit (&surface->base);
  
  free (surface);
}

static void
_glc_agl_surface_realize (void *abstract_surface)
{
  glc_agl_surface_t *surface = (glc_agl_surface_t *) abstract_surface;
  
  if (surface->window) {
    _glc_agl_surface_realize_window (surface->window,
                                     &surface->base.width,
                                     &surface->base.height);
    
    glc_agl_context_push_current (surface, GLC_CN_SURFACE_DRAWABLE_CURRENT);
    
    aglUpdateContext (surface->context->context);
    
    glc_agl_context_pop_current (surface);
  }
}

static void
_glc_agl_surface_show (void *abstract_surface)
{
  glc_agl_surface_t *surface = (glc_agl_surface_t *) abstract_surface;
  
  if (!surface->window)
    return;

  glc_agl_context_push_current (surface, GLC_CN_SURFACE_DRAWABLE_CURRENT);
  
  aglSwapBuffers (surface->context->context);
  
  glc_agl_context_pop_current (surface);
}
