/*
 * Copyright  2004 David Reveman, Peter Nilsson
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the names of
 * David Reveman and Peter Nilsson not be used in advertising or
 * publicity pertaining to distribution of the software without
 * specific, written prior permission. David Reveman and Peter Nilsson
 * makes no representations about the suitability of this software for
 * any purpose. It is provided "as is" without express or implied warranty.
 *
 * DAVID REVEMAN AND PETER NILSSON DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL DAVID REVEMAN AND
 * PETER NILSSON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: David Reveman <c99drn@cs.umu.se>
 *          Peter Nilsson <c99pnn@cs.umu.se>
 */

#ifdef HAVE_CONFIG_H
#  include "../config.h"
#endif

#include "glcint.h"

#include <dlfcn.h>

glc_glx_static_proc_address_list_t _glc_glx_proc_address = {
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  0
};

typedef void *(* glc_proc_ptr_get_proc_address_arb_t)(GLubyte *);

glc_proc_ptr_get_proc_address_arb_t glc_glx_get_proc_address_arb = NULL;

void *
glc_glx_get_proc_address (const char *name)
{
  void *address = NULL;
  
  if (glc_glx_get_proc_address_arb)
    address = glc_glx_get_proc_address_arb ((GLubyte *) name);
  
  if (!address) {
    void *dlhand;
    
    if ((dlhand = dlopen (NULL, RTLD_LAZY))) {
      address = dlsym (dlhand, name);
      dlclose (dlhand);
    }
  }
  
  return address;
}

static void
glc_glx_proc_address_lookup (void)
{
  _glc_glx_proc_address.supported = 1;
  glc_glx_get_proc_address_arb =
    (glc_proc_ptr_get_proc_address_arb_t)
    glc_glx_get_proc_address ("glXGetProcAddressARB");
  _glc_glx_proc_address.glx_get_fbconfigs =
    (glc_proc_ptr_glx_get_fbconfigs_t)
    glc_glx_get_proc_address ("glXGetFBConfigs");
  _glc_glx_proc_address.glx_get_fbconfig_attrib =
    (glc_proc_ptr_glx_get_fbconfig_attrib_t)
    glc_glx_get_proc_address ("glXGetFBConfigAttrib");
  _glc_glx_proc_address.glx_get_visual_from_fbconfig =
    (glc_proc_ptr_glx_get_visual_from_fbconfig_t)
    glc_glx_get_proc_address ("glXGetVisualFromFBConfig");
  _glc_glx_proc_address.glx_create_pbuffer =
    (glc_proc_ptr_glx_create_pbuffer_t)
    glc_glx_get_proc_address ("glXCreatePbuffer");
  _glc_glx_proc_address.glx_destroy_pbuffer =
    (glc_proc_ptr_glx_destroy_pbuffer_t)
    glc_glx_get_proc_address ("glXDestroyPbuffer");
}

#ifdef XTHREADS

#include <X11/Xthreads.h>
#include <stdlib.h>

/* thread safe */
static int tsd_initialized = 0;
static xthread_key_t info_tsd;

glc_glx_thread_info_t *
glc_glx_thread_info_get (void)
{
  if (!tsd_initialized) {
    glc_glx_thread_info_t *info = (glc_glx_thread_info_t *)
      malloc (sizeof (glc_glx_thread_info_t));
    info->displays = NULL;
    info->n_displays = 0;
    glc_glx_proc_address_lookup ();
    xthread_key_create (&info_tsd, NULL);
    xthread_set_specific (info_tsd, info);
    tsd_initialized = 1;
    return info;
  } else {
    void *p;

    xthread_get_specific (info_tsd, &p);
    return (glc_glx_thread_info_t *) p;
  }
}

#else

/* not thread safe */
static glc_glx_thread_info_t thread_info = {
  NULL,
  0
};

glc_glx_thread_info_t *
glc_glx_thread_info_get (void)
{
  return &thread_info;
}

#endif

glc_glx_display_info_t *
glc_glx_display_info_get (Display *display)
{
  glc_glx_display_info_t *display_info;
  glc_glx_thread_info_t *thread_info = glc_glx_thread_info_get ();
  glc_glx_display_info_t **displays = thread_info->displays;
  int index, n_displays = thread_info->n_displays;

  for (; n_displays; n_displays--, displays++)
    if ((*displays)->display == display)
      return *displays;

  index = thread_info->n_displays++;

  thread_info->displays =
    realloc (thread_info->displays,
             sizeof (glc_glx_display_info_t *) * thread_info->n_displays);

  display_info = malloc (sizeof (glc_glx_display_info_t));
  thread_info->displays[index] = display_info;
  
  display_info->thread_info = thread_info;
  display_info->display = display;
  display_info->screens = NULL;
  display_info->n_screens = 0;
  
  return display_info;
}

static void
glc_glx_create_root_context (glc_glx_screen_info_t *screen_info)
{
  XVisualInfo *vinfo;
  XSetWindowAttributes win_attrib;
  int attrib[] = {
    GLX_RGBA,
    0
  };
  int screen = screen_info->screen;
  Display *display = screen_info->display_info->display;
  
  vinfo = glXChooseVisual (display, screen, attrib);
  if (!vinfo)
    return;
  
  win_attrib.colormap =
    XCreateColormap (display,
                     RootWindow (display, screen),
                     vinfo->visual, AllocNone);
  screen_info->root_drawable =
    XCreateWindow (display,
                   RootWindow (display, screen),
                   -2, -2, 1, 1, 0, vinfo->depth, CopyFromParent,
                   vinfo->visual, CWColormap, &win_attrib);

  screen_info->root_context.context =
    glXCreateContext (display, vinfo, NULL, 1);

  screen_info->root_context.fbconfig = (XID) 0;  
  screen_info->root_context.id = vinfo->visualid;

  memset (&screen_info->root_context.gl_proc_address, 0,
          sizeof (glc_gl_proc_address_list_t));
  memset (&screen_info->root_context.glx_proc_address, 0,
          sizeof (glc_glx_proc_address_list_t));
  
  if (_glc_glx_proc_address.supported) {
    screen_info->root_context.gl_proc_address.supported =
      screen_info->root_context.glx_proc_address.supported = 1;
  }
  
  XFree (vinfo);
}

glc_glx_screen_info_t *
glc_glx_screen_info_get (Display *display,
                         int screen)
{
  glc_glx_screen_info_t *screen_info;
  glc_glx_display_info_t *display_info = glc_glx_display_info_get (display);
  glc_glx_screen_info_t **screens = display_info->screens;
  int index, n_screens = display_info->n_screens;

  for (; n_screens; n_screens--, screens++)
    if ((*screens)->screen == screen)
      return *screens;

  index = display_info->n_screens++;

  display_info->screens =
    realloc (display_info->screens,
             sizeof (glc_glx_screen_info_t *) * display_info->n_screens);

  screen_info = malloc (sizeof (glc_glx_screen_info_t));
  display_info->screens[index] = screen_info;

  screen_info->display_info = display_info;
  screen_info->screen = screen;
  screen_info->formats = NULL;
  screen_info->format_ids = NULL;
  screen_info->n_formats = 0;

  screen_info->contexts = NULL;
  screen_info->n_contexts = 0;

  memset (&screen_info->programs, 0, sizeof (glc_programs_t));

  glc_glx_create_root_context (screen_info);

  glXMakeCurrent (screen_info->display_info->display,
                  screen_info->root_drawable,
                  screen_info->root_context.context);
  glPixelStorei (GL_PACK_ALIGNMENT, 4);
  glPixelStorei (GL_UNPACK_ALIGNMENT, 4);

  glc_glx_query_extensions (screen_info);
  glc_glx_query_formats (screen_info);

  if (screen_info->root_context.gl_proc_address.supported)
    glc_glx_context_proc_address_lookup (&screen_info->root_context);
  
  screen_info->context_stack = malloc (sizeof (glc_glx_context_info_t));
  screen_info->context_stack_size = 1;
  screen_info->context_stack->surface = NULL;
  screen_info->context_stack->constraint = GLC_CN_NONE;

  return screen_info;
}

void
glc_glx_initialize (Display *display,
                    int screen)
{
  glc_glx_screen_info_get (display, screen);
}
slim_hidden_def(glc_glx_initialize);

unsigned long
glc_glx_features (Display *display,
                  int screen)
{
  glc_glx_screen_info_t *screen_info =
    glc_glx_screen_info_get (display, screen);

  return screen_info->feature_mask;
}
slim_hidden_def(glc_glx_features);
