/*
 * Copyright  2004 David Reveman, Peter Nilsson
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the names of
 * David Reveman and Peter Nilsson not be used in advertising or
 * publicity pertaining to distribution of the software without
 * specific, written prior permission. David Reveman and Peter Nilsson
 * makes no representations about the suitability of this software for
 * any purpose. It is provided "as is" without express or implied warranty.
 *
 * DAVID REVEMAN AND PETER NILSSON DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL DAVID REVEMAN AND
 * PETER NILSSON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: David Reveman <c99drn@cs.umu.se>
 *          Peter Nilsson <c99pnn@cs.umu.se>
 */

#ifdef HAVE_CONFIG_H
#  include "../config.h"
#endif

#include "glcint.h"

static glc_surface_t *
_glc_glx_surface_create_similar (void *abstract_templ,
                                 glc_format_name_t format_name,
                                 int width,
                                 int height);

static void
_glc_glx_surface_destroy (void *abstract_surface);

static glc_texture_t *
_glc_glx_surface_get_texture (void *abstract_surface);

static void
_glc_glx_surface_realize (void *abstract_surface);

static void
_glc_glx_surface_show (void *abstract_surface);

static glc_bool_t
_glc_glx_surface_push_current (void *abstract_surface,
                               glc_constraint_t constraint)
{
  glc_glx_surface_t *surface = (glc_glx_surface_t *) abstract_surface;

  if (constraint == GLC_CN_SURFACE_DRAWABLE_CURRENT) {
    if (surface->render_texture) {
      surface->context->glx_proc_address.glx_release_tex_image_ati
        (surface->screen_info->display_info->display, surface->pbuffer,
         (surface->base.format->doublebuffer)?
         GLX_BACK_LEFT_ATI: GLX_FRONT_LEFT_ATI);
    }
    if (!surface->drawable)
      constraint = GLC_CN_ANY_CONTEXT_CURRENT;
  }
  
  surface = glc_glx_context_push_current (surface, constraint);

  if (surface) {
    glc_surface_setup_environment (&surface->base);
    return 1;
  }

  return 0;
}

static void
_glc_glx_surface_pop_current (void *abstract_surface)
{
  glc_glx_surface_t *surface = (glc_glx_surface_t *) abstract_surface;

  surface = glc_glx_context_pop_current (surface);
  
  if (surface)
    glc_surface_setup_environment (&surface->base);
}

static const struct glc_surface_backend glc_glx_surface_backend = {
  _glc_glx_surface_create_similar,
  _glc_glx_surface_destroy,
  _glc_glx_surface_push_current,
  _glc_glx_surface_pop_current,
  _glc_glx_surface_get_texture,
  _glc_glx_surface_realize,
  _glc_glx_surface_show
};

static glc_bool_t
_glc_glx_surface_realize_window (Display *display,
                                 Window drawable,
                                 int *width,
                                 int *height)
{
  unsigned int uwidth, uheight, bwidth_ignore, depth_ignore;
  Window root_ignore;
  int x_ignore, y_ignore;
  
  if (!XGetGeometry (display, drawable, &root_ignore, &x_ignore, &y_ignore,
                     &uwidth, &uheight, &bwidth_ignore, &depth_ignore))
    return 0;
  
  *width = (int) uwidth;
  *height = (int) uheight;
  
  return 1;
}

static void
_glc_glx_surface_ensure_texture (glc_glx_surface_t *surface)
{
  if (!surface->base.dirty)
    return;

  if (surface->render_texture) {
    glc_texture_bind (surface->base.texture);
    surface->context->glx_proc_address.glx_bind_tex_image_ati
      (surface->screen_info->display_info->display, surface->pbuffer,
       (surface->base.format->doublebuffer)?
       GLX_BACK_LEFT_ATI: GLX_FRONT_LEFT_ATI);
    
    glc_texture_unbind (surface->base.texture);

  } else
    glc_texture_copy_surface (surface->base.texture, &surface->base,
                              &surface->base.dirty_region);
  
  surface->base.dirty = 0;
}

static glc_texture_t *
_glc_glx_surface_get_texture (void *abstract_surface)
{
  glc_glx_surface_t *surface = (glc_glx_surface_t *) abstract_surface;

  if (!surface->base.texture->allocated)
    glc_texture_allocate (surface->base.texture);
  
  _glc_glx_surface_ensure_texture (surface);
  
  return surface->base.texture;
}

static void
_glc_glx_set_features (glc_glx_surface_t *surface)
{
  surface->base.feature_mask = surface->screen_info->feature_mask;

  surface->base.feature_mask &= ~GLC_FEATURE_ATI_RENDER_TEXTURE_MASK;
  surface->base.feature_mask &= ~GLC_FEATURE_ARB_VERTEX_PROGRAM_MASK;
  surface->base.feature_mask &= ~GLC_FEATURE_ARB_FRAGMENT_PROGRAM_MASK;
  surface->base.feature_mask &= ~GLC_FEATURE_CONVOLUTION_FILTER_MASK;
  surface->base.feature_mask &= ~GLC_FEATURE_GRADIENT_MASK;
  surface->base.feature_mask &= ~GLC_FEATURE_MULTISAMPLE_MASK;
  surface->base.feature_mask &= ~GLC_FEATURE_OFFSCREEN_MULTISAMPLE_MASK;

  if (surface->context->glx_proc_address.supported) {
    glc_surface_push_current (&surface->base, GLC_CN_SURFACE_CONTEXT_CURRENT);
    glc_surface_pop_current (&surface->base);
  }
  
  if ((surface->screen_info->feature_mask &
       GLC_FEATURE_ATI_RENDER_TEXTURE_MASK) &&
      surface->context->glx_proc_address.glx_bind_tex_image_ati &&
      surface->context->glx_proc_address.glx_release_tex_image_ati)
    surface->base.feature_mask |= GLC_FEATURE_ATI_RENDER_TEXTURE_MASK;

  if (surface->context->gl_proc_address.gl_active_texture_arb &&
      surface->context->gl_proc_address.gl_multi_tex_coord_2d_arb &&
      surface->context->gl_proc_address.gl_gen_programs_arb &&
      surface->context->gl_proc_address.gl_delete_programs_arb &&
      surface->context->gl_proc_address.gl_program_string_arb &&
      surface->context->gl_proc_address.gl_bind_program_arb) {
    if (surface->screen_info->feature_mask &
        GLC_FEATURE_ARB_FRAGMENT_PROGRAM_MASK) {
      surface->base.feature_mask |= GLC_FEATURE_ARB_FRAGMENT_PROGRAM_MASK;
      surface->base.feature_mask |= GLC_FEATURE_GRADIENT_MASK;
    }
    
    if (surface->screen_info->feature_mask &
        GLC_FEATURE_ARB_VERTEX_PROGRAM_MASK)
      surface->base.feature_mask |= GLC_FEATURE_ARB_VERTEX_PROGRAM_MASK;
    
    if ((surface->base.feature_mask & GLC_FEATURE_ARB_VERTEX_PROGRAM_MASK) &&
        (surface->base.feature_mask & GLC_FEATURE_ARB_FRAGMENT_PROGRAM_MASK) &&
        surface->context->gl_proc_address.gl_program_local_param_4d_arb &&
        surface->context->gl_proc_address.gl_get_program_iv_arb) {
      GLuint texture_indirections;

      surface->context->gl_proc_address.gl_get_program_iv_arb
        (GL_FRAGMENT_PROGRAM_ARB,
         GL_MAX_PROGRAM_TEX_INDIRECTIONS_ARB,
         &texture_indirections);

      /* Convolution filter programs require support for at least six
         texture indirections. */
      if (texture_indirections >= 5)
        surface->base.feature_mask |= GLC_FEATURE_CONVOLUTION_FILTER_MASK;
    }
  }

  if (surface->base.format->multisample.supported) {
    surface->base.feature_mask |= GLC_FEATURE_MULTISAMPLE_MASK;
    if (surface->pbuffer)
      surface->base.feature_mask |= GLC_FEATURE_OFFSCREEN_MULTISAMPLE_MASK;
  }
}

static glc_surface_t *
_glc_glx_surface_create (glc_glx_screen_info_t *screen_info,
                         glc_format_t *format,
                         int width,
                         int height)
{
  glc_glx_surface_t *surface;
  glc_glx_context_t *context;
  unsigned int texture_format;

  context = glc_glx_context_get (screen_info, format);
  if (!context)
    return NULL;
  
  surface = (glc_glx_surface_t *) calloc (1, sizeof (glc_glx_surface_t));
  if (surface == NULL)
    return NULL;

  glc_surface_init (&surface->base, &glc_glx_surface_backend);

  surface->screen_info = screen_info;
  surface->context = context;
  
  surface->base.programs = &screen_info->programs;
  surface->base.format = format;
  surface->base.width = width;
  surface->base.height = height;

  texture_format = glc_get_gl_format_from_bpp (format->bpp);

  if (screen_info->feature_mask & GLC_FEATURE_ATI_RENDER_TEXTURE_MASK)
    surface->render_texture = 1;
  
  glc_surface_push_current (&surface->base, GLC_CN_ANY_CONTEXT_CURRENT);

  surface->base.texture =
    glc_texture_generate (width, height,
                          texture_format,
                          screen_info->texture_mask);

  if (!surface->base.texture) {
    glc_surface_destroy (&surface->base);
    return NULL;
  }

  if (screen_info->feature_mask & GLC_FEATURE_OFFSCREEN_DRAWING_MASK)
    surface->drawable = surface->pbuffer =
      glc_glx_pbuffer_create (screen_info->display_info->display,
                              surface->context->fbconfig,
                              surface->base.texture,
                              surface->render_texture);

  _glc_glx_set_features (surface);
  
  surface->base.proc_address = &context->gl_proc_address;

  if (surface->base.feature_mask & GLC_FEATURE_ATI_RENDER_TEXTURE_MASK) {
    if (format->red_size || format->green_size || format->blue_size)
      surface->base.requires_flipping = 0;
  } else
    surface->render_texture = 0;
  
  if ((!surface->render_texture) && (!surface->pbuffer))
    glc_texture_allocate (surface->base.texture);

  glc_surface_pop_current (&surface->base);

  return &surface->base;
}

glc_surface_t *
glc_glx_surface_create (Display *display,
                        int screen,
                        glc_format_t *format,
                        int width,
                        int height)
{

  return
    _glc_glx_surface_create (glc_glx_screen_info_get
                             (display, screen),
                             format, width, height);
}
slim_hidden_def(glc_glx_surface_create);

glc_surface_t *
glc_glx_surface_create_for_window (Display *display,
                                   int screen,
                                   glc_format_t *format,
                                   Window window)
{
  glc_glx_surface_t *surface;
  glc_glx_context_t *context;
  int width, height;
  glc_glx_screen_info_t *screen_info =
    glc_glx_screen_info_get (display, screen);
  
  context = glc_glx_context_get (screen_info, format);
  if (!context)
    return NULL;

  if (!_glc_glx_surface_realize_window (display, window, &width, &height))
    return NULL;

  surface = (glc_glx_surface_t *) calloc (1, sizeof (glc_glx_surface_t));
  if (surface == NULL)
    return NULL;

  glc_surface_init (&surface->base, &glc_glx_surface_backend);
  
  surface->screen_info = screen_info;
  surface->context = context;

  surface->base.programs = &screen_info->programs;
  surface->base.format = format;
  surface->base.width = width;
  surface->base.height = height;
  
  surface->drawable = window;

  _glc_glx_set_features (surface);

  surface->base.proc_address = &context->gl_proc_address;
  
  return &surface->base;
}
slim_hidden_def(glc_glx_surface_create_for_window);

static glc_surface_t *
_glc_glx_surface_create_similar (void *abstract_templ,
                                 glc_format_name_t format_name,
                                 int width,
                                 int height)
{
  glc_glx_surface_t *templ = (glc_glx_surface_t *) abstract_templ;
  
  if (templ->screen_info->feature_mask & GLC_FEATURE_OFFSCREEN_DRAWING_MASK) {
    glc_format_t *format;
    
    format = glc_format_find_sufficient_standard
      (templ->base.format, 1, GLC_FORMAT_OPTION_OFFSCREEN_MASK, format_name);

    if (!format)
      format = glc_format_find_standard (templ->screen_info->formats,
                                         templ->screen_info->n_formats,
                                         GLC_FORMAT_OPTION_OFFSCREEN_MASK,
                                         format_name);
    if (format)
      return _glc_glx_surface_create (templ->screen_info, format,
                                      width, height);
  }
  
  return NULL;
}

static void
_glc_glx_surface_destroy (void *abstract_surface)
{
  glc_glx_surface_t *surface = (glc_glx_surface_t *) abstract_surface;

  glc_surface_push_current (&surface->base, GLC_CN_ANY_CONTEXT_CURRENT);

  if (surface->render_texture) {
    surface->context->glx_proc_address.glx_release_tex_image_ati
      (surface->screen_info->display_info->display, surface->pbuffer,
       (surface->base.format->doublebuffer)?
       GLX_BACK_LEFT_ATI: GLX_FRONT_LEFT_ATI);
    }
  
  if (surface->base.texture)
    glc_texture_destroy (surface->base.texture);

  if (surface->pbuffer)
    glc_glx_pbuffer_destroy (surface->screen_info->display_info->display,
                             surface->pbuffer);
  
  glc_surface_pop_current (&surface->base);

  if (glXGetCurrentDrawable () == surface->drawable) {
    surface->drawable = None;
    glc_glx_context_make_current (surface);
  }
  
  glc_surface_deinit (&surface->base);
  
  free (surface);
}


static void
_glc_glx_surface_realize (void *abstract_surface)
{
  glc_glx_surface_t *surface = (glc_glx_surface_t *) abstract_surface;
  
  if ((! surface->pbuffer) && surface->drawable) {
    _glc_glx_surface_realize_window
      (surface->screen_info->display_info->display, surface->drawable,
       &surface->base.width, &surface->base.height);
  }
}

static void
_glc_glx_surface_show (void *abstract_surface)
{
  glc_glx_surface_t *surface = (glc_glx_surface_t *) abstract_surface;

  if (surface->pbuffer || (!surface->drawable))
    return;

  glc_glx_context_push_current (surface, GLC_CN_SURFACE_DRAWABLE_CURRENT);
  
  glXSwapBuffers (surface->screen_info->display_info->display,
                  surface->drawable);
  
  glc_glx_context_pop_current (surface);
}
