/*
 * Copyright  2004 David Reveman, Peter Nilsson
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the names of
 * David Reveman and Peter Nilsson not be used in advertising or
 * publicity pertaining to distribution of the software without
 * specific, written prior permission. David Reveman and Peter Nilsson
 * makes no representations about the suitability of this software for
 * any purpose. It is provided "as is" without express or implied warranty.
 *
 * DAVID REVEMAN AND PETER NILSSON DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL DAVID REVEMAN AND
 * PETER NILSSON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: David Reveman <c99drn@cs.umu.se>
 *          Peter Nilsson <c99pnn@cs.umu.se>
 */

#ifdef HAVE_CONFIG_H
#  include "../config.h"
#endif

#include "glcint.h"

#include <stdio.h>

#define EXPAND_NONE ""
#define EXPAND_2D_TEX "2D"
#define EXPAND_RECT_TEX "RECT"
#define EXPAND_NO_PD_OP \
  "MOV result.color, color;\n"

#define EXPAND_SRC_TEMP "TEMP src;\n"
#define EXPAND_SRC_2D_PD_OP \
  "TEX src, fragment.texcoord[0], texture[0], 2D;\n" \
  "MUL result.color, src, color.a;\n"
#define EXPAND_SRC_RECT_PD_OP \
  "TEX src, fragment.texcoord[0], texture[0], RECT;\n" \
  "MUL result.color, src, color.a;\n"

#define EXPAND_MASK_TEMP "TEMP mask;\n"
#define EXPAND_MASK_2D_PD_OP \
  "TEX mask, fragment.texcoord[1], texture[1], 2D;\n" \
  "MUL result.color, color, mask.a;\n"
#define EXPAND_MASK_RECT_PD_OP \
  "TEX mask, fragment.texcoord[1], texture[1], RECT;\n" \
  "MUL result.color, color, mask.a;\n"

typedef struct _glc_program_expand_t glc_program_expand_t;

static const struct _glc_program_expand_t {
  int index;
  char *tex;
  char *temporary;
  char *operation;
} _program_expand_map[] = {

  /* src is operation surface */

  /* GLC_PROGRAM_2DSRC_2DMASK_OFFSET */
  { 0, EXPAND_2D_TEX, EXPAND_MASK_TEMP, EXPAND_MASK_2D_PD_OP },

  /* GLC_PROGRAM_RECTSRC_2DMASK_OFFSET */
  { 0, EXPAND_RECT_TEX, EXPAND_MASK_TEMP, EXPAND_MASK_2D_PD_OP },

  /* GLC_PROGRAM_2DSRC_RECTMASK_OFFSET */
  { 0, EXPAND_2D_TEX, EXPAND_MASK_TEMP, EXPAND_MASK_RECT_PD_OP },

  /* GLC_PROGRAM_RECTSRC_RECTMASK_OFFSET */
  { 0, EXPAND_RECT_TEX, EXPAND_MASK_TEMP, EXPAND_MASK_RECT_PD_OP },

  /* GLC_PROGRAM_NOSRC_2DMASK_OFFSET */
  { 0, EXPAND_NONE, EXPAND_MASK_TEMP, EXPAND_MASK_2D_PD_OP },

  /* GLC_PROGRAM_NOSRC_RECTMASK_OFFSET */
  { 0, EXPAND_NONE, EXPAND_MASK_TEMP, EXPAND_MASK_RECT_PD_OP },

  /* GLC_PROGRAM_2DSRC_NOMASK_OFFSET */
  { 0, EXPAND_2D_TEX, EXPAND_NONE, EXPAND_NO_PD_OP },

  /* GLC_PROGRAM_RECTSRC_NOMASK_OFFSET */
  { 0, EXPAND_RECT_TEX, EXPAND_NONE, EXPAND_NO_PD_OP },

  /* GLC_PROGRAM_NOSRC_NOMASK_OFFSET */
  { 0, EXPAND_NONE, EXPAND_NONE, EXPAND_NO_PD_OP },

  
  /* mask is operation surface */

  /* GLC_PROGRAM_2DSRC_2DMASK_OFFSET */
  { 1, EXPAND_2D_TEX, EXPAND_SRC_TEMP, EXPAND_SRC_2D_PD_OP },

  /* GLC_PROGRAM_RECTSRC_2DMASK_OFFSET */
  { 1, EXPAND_2D_TEX, EXPAND_SRC_TEMP, EXPAND_SRC_RECT_PD_OP },

  /* GLC_PROGRAM_2DSRC_RECTMASK_OFFSET */
  { 1, EXPAND_RECT_TEX, EXPAND_SRC_TEMP, EXPAND_SRC_2D_PD_OP },

  /* GLC_PROGRAM_RECTSRC_RECTMASK_OFFSET */
  { 1, EXPAND_RECT_TEX, EXPAND_SRC_TEMP, EXPAND_SRC_RECT_PD_OP },

  /* GLC_PROGRAM_NOSRC_2DMASK_OFFSET */
  { 1, EXPAND_2D_TEX, EXPAND_NONE, EXPAND_NO_PD_OP },

  /* GLC_PROGRAM_NOSRC_RECTMASK_OFFSET */
  { 1, EXPAND_RECT_TEX, EXPAND_NONE, EXPAND_NO_PD_OP },

  /* GLC_PROGRAM_2DSRC_NOMASK_OFFSET */
  { 1, EXPAND_NONE, EXPAND_SRC_TEMP, EXPAND_SRC_2D_PD_OP },

  /* GLC_PROGRAM_RECTSRC_NOMASK_OFFSET */
  { 1, EXPAND_NONE, EXPAND_SRC_TEMP, EXPAND_SRC_RECT_PD_OP },

  /* GLC_PROGRAM_NOSRC_NOMASK_OFFSET */
  { 1, EXPAND_NONE, EXPAND_NONE, EXPAND_NO_PD_OP }
};

/*
 * Passes texture coordinates to convolution filter
 * fragment programs.
 *
 * program.local[0]: Vertical pixel offset in texture coordinates
 * program.local[1]: Horizontal pixel offset in texture coordinates
 *
 * Author: David Reveman <c99drn@cs.umu.se>
 */   
static char *_glc_vertex_program_convolution =
"!!ARBvp1.0\n"
"OPTION ARB_position_invariant;\n"
"ATTRIB coord = vertex.texcoord[%d];\n"
"PARAM vertical_offset   = program.local[0];\n"
"PARAM horizontal_offset = program.local[1];\n"
"MOV result.texcoord[%d], coord;\n"
"ADD result.texcoord[2], coord, vertical_offset;\n"
"SUB result.texcoord[3], coord, vertical_offset;\n"
"ADD result.texcoord[4], coord, horizontal_offset;\n"
"SUB result.texcoord[5], coord, horizontal_offset;\n"
"MOV result.texcoord[%d], vertex.texcoord[%d];\n"
"END";

/*
 * Porter-Duff compositing (SRC in MASK).
 * Texture unit 0 is SRC.
 * Texture unit 1 is MASK.
 *
 * Author: David Reveman <c99drn@cs.umu.se>
 */
static char *_glc_fragment_program_simple =
"!!ARBfp1.0\n"
"TEMP color;\n"

/* temporary */
"%s"

/* src texture */
"TEX color, fragment.texcoord[0], texture[0], %s;\n"

/* pd operation */
"%s"

"END";

/*
 * 3x3 convolution filter.
 * Convolution kernel must be normalized.
 *
 * program.local[0]: Top convolution kernel row
 * program.local[1]: Middle convolution kernel row
 * program.local[2]: Bottom convolution kernel row
 *
 * Author: David Reveman <c99drn@cs.umu.se>
 */
static char *_glc_fragment_program_convolution =
"!!ARBfp1.0\n"
"ATTRIB east = fragment.texcoord[2];\n"
"ATTRIB west = fragment.texcoord[3];\n"
"ATTRIB south = fragment.texcoord[4];\n"
"ATTRIB north = fragment.texcoord[5];\n"
"PARAM k0 = program.local[0];\n"
"PARAM k1 = program.local[1];\n"
"PARAM k2 = program.local[2];\n"
"TEMP color, in, coord;\n"

/* temporary */
"%s"

/* center */
"TEX in, fragment.texcoord[%d], texture[%d], %s;\n"
"MUL color, in, k1.y;\n"

/* north west */
"MOV coord.x, west.x;\n"
"MOV coord.y, north.y;\n"
"TEX in, coord, texture[%d], %s;\n"
"MAD color, in, k0.x, color;\n"

/* north */
"TEX in, north, texture[%d], %s;\n"
"MAD color, in, k0.y, color;\n"

/* north east */
"MOV coord.x, east.x;\n"
"TEX in, coord, texture[%d], %s;\n"
"MAD color, in, k0.z, color;\n"

/* east */
"TEX in, east, texture[%d], %s;\n"
"MAD color, in, k1.x, color;\n"

/* south east */
"MOV coord.y, south.y;\n"
"TEX in, coord, texture[%d], %s;\n"
"MAD color, in, k2.z, color;\n"

/* south */
"TEX in, south, texture[%d], %s;\n"
"MAD color, in, k2.y, color;\n"

/* south west */
"MOV coord.x, west.x;\n"
"TEX in, coord, texture[%d], %s;\n"
"MAD color, in, k2.x, color;\n"

/* west */
"TEX in, west, texture[%d], %s;\n"
"MAD color, in, k1.x, color;\n"

/* pd operation */
"%s"

"END";

char *_glc_fragment_program_programmatic[] = {
  /*
   * Solid.
   *
   * fragment.color: color
   *
   * Author: David Reveman <c99drn@cs.umu.se>
   */
  "!!ARBfp1.0\n"
  "ATTRIB color = fragment.color;\n"

  /* temporary */
  "%s"

  /* pd operation */
  "%s"

  "END",
  
  /*
   * Linear gradient.
   *
   * program.local[0]: start color
   * program.local[1]: end color
   * program.local[2].x = start offset
   * program.local[2].y = stop offset
   * program.local[2].z = 1 / length
   * program.local[3].x = sin (angle)
   * program.local[3].y = cos (angle)
   *
   * Author: David Reveman <c99drn@cs.umu.se>
   */
  "!!ARBfp1.0\n"
  "PARAM start_color = program.local[0];\n"
  "PARAM end_color = program.local[1];\n"
  "PARAM gradient = program.local[2];\n"
  "PARAM angle = program.local[3];\n"
  "ATTRIB pos = fragment.texcoord[%d];\n"
  "TEMP color, distance, position;\n"

  /* temporary */
  "%s"

  "MUL position.x, angle.x, pos.x;\n"
  "MAD position.x, angle.y, pos.y, position.x;\n"

  "SUB distance, position.x, gradient.x;\n"
  "MUL distance, distance, gradient.z;\n"

  "MUL color, distance, end_color;\n"

  "SUB distance, gradient.y, position.x;\n"
  "MUL distance, distance, gradient.z;\n"

  "MAD color, distance, start_color, color;\n"

  /* pd operation */
  "%s"
  
  "END",

  /*
   * Radial gradient.
   *
   * program.local[0]: start color
   * program.local[1]: end color
   * program.local[2].x = center point X coordinate
   * program.local[2].y = center point Y coordinate
   * program.local[2].z = 1 / (length * length)
   *
   * Author: David Reveman <c99drn@cs.umu.se>
   */

  "!!ARBfp1.0\n"
  "PARAM start_color = program.local[0];\n"
  "PARAM end_color = program.local[1];\n"
  "PARAM gradient = program.local[2];\n"
  "ATTRIB position = fragment.texcoord[%d];\n"
  "TEMP color, distance;\n"

  /* temporary */
  "%s"

  "MOV distance, 0.0;\n"
  "SUB distance.x, position.x, gradient.x;\n"
  "SUB distance.y, position.y, gradient.y;\n"
  
  "DP3 distance, distance, distance;\n"
  "MUL distance.xyzw, distance, gradient.z;\n"
  
  "MUL color, distance, end_color;\n"
  "SUB distance, 1.0, distance;\n"
  "MAD color, distance, start_color, color;\n"

  /* pd operation */
  "%s"
  
  "END"
};

static unsigned int
glc_program_compile_vertex_arb (glc_gl_proc_address_list_t *proc_address,
                                char *program_string)
{
  GLint error;
  GLuint program;
    
  proc_address->gl_gen_programs_arb (1, (GLuint *) &program);
  proc_address->gl_bind_program_arb (GL_VERTEX_PROGRAM_ARB, program);
  proc_address->gl_program_string_arb (GL_VERTEX_PROGRAM_ARB,
                                       GL_PROGRAM_FORMAT_ASCII_ARB,
                                       strlen (program_string),
                                       program_string);
  
  glGetIntegerv (GL_PROGRAM_ERROR_POSITION_ARB, &error);
  if (error != -1) {
    proc_address->gl_delete_programs_arb (1, &program);
    program = 0;
  }
  
  return (unsigned int) program;
}

static unsigned int
glc_program_compile_fragment_arb (glc_gl_proc_address_list_t *proc_address,
                                  char *program_string)
{
  GLint error;
  GLuint program;
    
  proc_address->gl_gen_programs_arb (1, (GLuint *) &program);
  proc_address->gl_bind_program_arb (GL_FRAGMENT_PROGRAM_ARB, program);
  proc_address->gl_program_string_arb (GL_FRAGMENT_PROGRAM_ARB,
                                       GL_PROGRAM_FORMAT_ASCII_ARB,
                                       strlen (program_string),
                                       program_string);
  
  glGetIntegerv (GL_PROGRAM_ERROR_POSITION_ARB, &error);
  if (error != -1) {
    proc_address->gl_delete_programs_arb (1, &program);
    program = 0;
  }
  
  return (unsigned int) program;
}

static int
_glc_program_offset (glc_texture_t *src_texture,
                     glc_texture_t *mask_texture)
{
  int offset;

  if (src_texture) {
    if (src_texture->target == GL_TEXTURE_2D) {
      offset = GLC_PROGRAM_2DSRC_NOMASK_OFFSET;
      
      if (mask_texture)
        offset = (mask_texture->target == GL_TEXTURE_2D)?
          GLC_PROGRAM_2DSRC_2DMASK_OFFSET: GLC_PROGRAM_2DSRC_RECTMASK_OFFSET;
    } else {
      offset = GLC_PROGRAM_RECTSRC_NOMASK_OFFSET;
      
      if (mask_texture)
        offset = (mask_texture->target == GL_TEXTURE_2D)?
          GLC_PROGRAM_RECTSRC_2DMASK_OFFSET:
      GLC_PROGRAM_RECTSRC_RECTMASK_OFFSET;
    }
  } else {
    offset = GLC_PROGRAM_NOSRC_NOMASK_OFFSET;
    
    if (mask_texture)
      offset = (mask_texture->target == GL_TEXTURE_2D)?
        GLC_PROGRAM_NOSRC_2DMASK_OFFSET: GLC_PROGRAM_NOSRC_RECTMASK_OFFSET;
  }
  
  return offset;
}

static unsigned long
glc_program_compile_simple (glc_gl_proc_address_list_t *proc_address,
                            int offset)
{
  char program_buffer[256];
  const glc_program_expand_t *expand = &_program_expand_map[offset];
  
  sprintf (program_buffer,
           _glc_fragment_program_simple,
           expand->temporary,
           expand->tex,
           expand->operation);

  return glc_program_compile_fragment_arb (proc_address, program_buffer);
}

static unsigned long
glc_program_compile_vertex_convolution
(glc_gl_proc_address_list_t *proc_address,
 int offset)
{
  char program_buffer[512];
  int conv_index = (offset)? 1: 0;
  int other_index = (offset)? 0: 1;
  
  sprintf (program_buffer,
           _glc_vertex_program_convolution,
           conv_index, conv_index,
           other_index, other_index);

  return glc_program_compile_vertex_arb (proc_address, program_buffer);
}

static unsigned long
glc_program_compile_convolution (glc_gl_proc_address_list_t *proc_address,
                                 int offset,
                                 int solid_offset)
{
  char *solid_op_table[] = {
    "MUL result.color, color, solid.a;\n",
    "MUL result.color, solid, color.a;\n",
  };
  char program_buffer[1280];
  const glc_program_expand_t *expand = &_program_expand_map[offset];
  char *temporary, *operation;

  if (solid_offset) {
    temporary = "PARAM solid = program.local[3];\n";
    operation = solid_op_table[solid_offset - 1];
  } else {
    temporary = expand->temporary;
    operation = expand->operation;
  }

  sprintf (program_buffer,
           _glc_fragment_program_convolution,
           temporary,
           expand->index, expand->index, expand->tex,
           expand->index, expand->tex,
           expand->index, expand->tex,
           expand->index, expand->tex,
           expand->index, expand->tex,
           expand->index, expand->tex,
           expand->index, expand->tex,
           expand->index, expand->tex,
           expand->index, expand->tex,
           operation);

  return glc_program_compile_fragment_arb (proc_address, program_buffer);
}

static unsigned long
glc_program_compile_programmatic (glc_programmatic_surface_type_t type,
                                  glc_gl_proc_address_list_t *proc_address,
                                  int offset)
{
  char program_buffer[1024];
  const glc_program_expand_t *expand = &_program_expand_map[offset];

  switch (type) {
  case GLC_PROGRAMMATIC_SURFACE_SOLID_TYPE:
    sprintf (program_buffer,
             _glc_fragment_program_programmatic[type],
             expand->temporary,
             expand->operation);
    break;
  case GLC_PROGRAMMATIC_SURFACE_LINEAR_GRADIENT_TYPE:
  case GLC_PROGRAMMATIC_SURFACE_RADIAL_GRADIENT_TYPE:
    sprintf (program_buffer,
             _glc_fragment_program_programmatic[type],
             expand->index,
             expand->temporary,
             expand->operation);
    break;
  }

  return glc_program_compile_fragment_arb (proc_address, program_buffer);
}

static void
glc_program_enable_simple (glc_gl_proc_address_list_t *proc_address,
                           glc_programs_t *programs,
                           glc_texture_t *src_texture,
                           glc_texture_t *mask_texture)
{
  int offset = _glc_program_offset (src_texture, mask_texture);
  
  if (!programs->fragment_simple[offset])
    programs->fragment_simple[offset] =
      glc_program_compile_simple (proc_address, offset);
  
  if (programs->fragment_simple[offset]) {
    glEnable (GL_FRAGMENT_PROGRAM_ARB);
    proc_address->gl_bind_program_arb (GL_FRAGMENT_PROGRAM_ARB,
                                       programs->fragment_simple[offset]);
  }
}

static void
glc_program_enable_convolution (glc_gl_proc_address_list_t *proc_address,
                                glc_programs_t *programs,
                                glc_surface_t *src,
                                glc_surface_t *mask,
                                glc_texture_t *src_texture,
                                glc_texture_t *mask_texture,
                                int offset,
                                int solid_offset)
{
  glc_texture_t *texture;
  glc_surface_t *surface;
  int fragment_offset, vertex_offset = (offset)? 1: 0;

  if (offset) {
    texture = mask_texture;
    surface = mask;
  } else {
    texture = src_texture;
    surface = src;
  }

  offset += _glc_program_offset (src_texture, mask_texture);

  fragment_offset = offset + GLC_FRAGMENT_PROGRAM_TYPES * solid_offset;

  if (!programs->vertex_convolution[vertex_offset])
    programs->vertex_convolution[vertex_offset] =
      glc_program_compile_vertex_convolution (proc_address, vertex_offset);
  
  if (!programs->fragment_convolution[fragment_offset])
    programs->fragment_convolution[fragment_offset] =
      glc_program_compile_convolution (proc_address, offset, solid_offset);
  
  if (programs->fragment_convolution[fragment_offset] &&
      programs->vertex_convolution[vertex_offset]) {

    glEnable (GL_VERTEX_PROGRAM_ARB);
    proc_address->gl_bind_program_arb
      (GL_VERTEX_PROGRAM_ARB, programs->vertex_convolution[vertex_offset]);
    proc_address->gl_program_local_param_4d_arb (GL_VERTEX_PROGRAM_ARB, 0,
                                                 texture->texcoord_width /
                                                 (double) texture->width,
                                                 0.000, 0.0, 0.0);
    proc_address->gl_program_local_param_4d_arb (GL_VERTEX_PROGRAM_ARB, 1,
                                                 0.000,
                                                 texture->texcoord_height /
                                                 (double) texture->height,
                                                 0.0, 0.0);
    
    glEnable (GL_FRAGMENT_PROGRAM_ARB);
    proc_address->gl_bind_program_arb
      (GL_FRAGMENT_PROGRAM_ARB,
       programs->fragment_convolution[fragment_offset]);
    proc_address->gl_program_local_param_4d_arb (GL_FRAGMENT_PROGRAM_ARB, 0,
                                                 surface->convolution->m[0][0],
                                                 surface->convolution->m[0][1],
                                                 surface->convolution->m[0][2],
                                                 0.0);
    proc_address->gl_program_local_param_4d_arb (GL_FRAGMENT_PROGRAM_ARB, 1,
                                                 surface->convolution->m[1][0],
                                                 surface->convolution->m[1][1],
                                                 surface->convolution->m[1][2],
                                                 0.0);
    proc_address->gl_program_local_param_4d_arb (GL_FRAGMENT_PROGRAM_ARB, 2,
                                                 surface->convolution->m[2][0],
                                                 surface->convolution->m[2][1],
                                                 surface->convolution->m[2][2],
                                                 0.0);

    if (solid_offset) {
      glc_solid_t *solid;
      
      if (solid_offset == 1)
        solid = &((glc_programmatic_surface_t *) mask)->attributes.solid;
      else
        solid = &((glc_programmatic_surface_t *) src)->attributes.solid;
      
      proc_address->gl_program_local_param_4d_arb
        (GL_FRAGMENT_PROGRAM_ARB, 3,
         (double) solid->color.red / 65536.0,
         (double) solid->color.green / 65536.0,
         (double) solid->color.blue / 65536.0,
         (double) solid->color.alpha / 65536.0);
    }
  }
}

static void
glc_program_enable_programmatic (glc_surface_t *dst,
                                 glc_programmatic_surface_t *surface,
                                 glc_texture_t *src_texture,
                                 glc_texture_t *mask_texture,
                                 int offset)
{
  int type_offset, add_offset;
  glc_gl_proc_address_list_t *proc_address = dst->proc_address;
  glc_programs_t *programs = dst->programs;

  if (offset)
    add_offset = _glc_program_offset (src_texture, NULL);
  else
    add_offset = _glc_program_offset (NULL, mask_texture);

  offset += add_offset;

  type_offset = offset + GLC_FRAGMENT_PROGRAM_TYPES * surface->type;

  /* no fragment proram needed for solid programmatic surface and no mask */
  if ((surface->type == GLC_PROGRAMMATIC_SURFACE_SOLID_TYPE) &&
      (add_offset == GLC_PROGRAM_NOSRC_NOMASK_OFFSET)) {
    glc_programmatic_surface_set_parameters (proc_address, surface);
    
    return;
  }

  if (dst->feature_mask & GLC_FEATURE_ARB_FRAGMENT_PROGRAM_MASK) {
    if (!programs->fragment_programmatic[type_offset])
      programs->fragment_programmatic[type_offset] =
        glc_program_compile_programmatic (surface->type, proc_address, offset);
    
    if (programs->fragment_programmatic[type_offset]) {
      glEnable (GL_FRAGMENT_PROGRAM_ARB);
      proc_address->gl_bind_program_arb
        (GL_FRAGMENT_PROGRAM_ARB,
         programs->fragment_programmatic[type_offset]);
    
      glc_programmatic_surface_set_parameters (proc_address, surface);
    }
  }
}

glc_program_type_t
glc_program_type (glc_surface_t *dst,
                  glc_surface_t *src,
                  glc_surface_t *mask)
{
  glc_program_type_t type = GLC_PROGRAM_TYPE_NONE;
  
  if (dst->feature_mask & GLC_FEATURE_ARB_FRAGMENT_PROGRAM_MASK) {

    if (dst->feature_mask & GLC_FEATURE_CONVOLUTION_FILTER_MASK)
    
    if (src->convolution) {
      if (mask && mask->programmatic &&
          ((glc_programmatic_surface_t *) mask)->type ==
          GLC_PROGRAMMATIC_SURFACE_SOLID_TYPE) {
        if (dst->feature_mask & GLC_FEATURE_CONVOLUTION_FILTER_MASK) {
          type = GLC_PROGRAM_TYPE_SRC_CONVOLUTION_AND_SOLID_MASK;
          goto OK2;
        } else {
          type = GLC_PROGRAM_TYPE_MASK_PROGRAMMATIC;
          goto OK1;
        }
      }
      type = GLC_PROGRAM_TYPE_SRC_CONVOLUTION;
      goto OK1;
    }
    
    if (mask && mask->convolution) {
      if (src->programmatic &&
          ((glc_programmatic_surface_t *) src)->type ==
          GLC_PROGRAMMATIC_SURFACE_SOLID_TYPE) {
        if (dst->feature_mask & GLC_FEATURE_CONVOLUTION_FILTER_MASK) {
          type = GLC_PROGRAM_TYPE_MASK_CONVOLUTION_AND_SOLID_SRC;
          goto OK2;
        } else {
          type = GLC_PROGRAM_TYPE_SRC_PROGRAMMATIC;
          goto OK1;
        }
      }
      type = GLC_PROGRAM_TYPE_MASK_CONVOLUTION;
      goto OK1;
    }

    if (src->programmatic) {
      type = GLC_PROGRAM_TYPE_SRC_PROGRAMMATIC;
      goto OK1;
    }
    
    if (mask && mask->programmatic) {
      type = GLC_PROGRAM_TYPE_MASK_PROGRAMMATIC;
      goto OK1;
    }      
  }
  
  if (src->programmatic &&
      ((glc_programmatic_surface_t *) src)->type ==
      GLC_PROGRAMMATIC_SURFACE_SOLID_TYPE) {
    type = GLC_PROGRAM_TYPE_SRC_PROGRAMMATIC;
    goto OK1;
  }
  
  if (mask && mask->programmatic &&
      ((glc_programmatic_surface_t *) mask)->type ==
      GLC_PROGRAMMATIC_SURFACE_SOLID_TYPE) {
    type = GLC_PROGRAM_TYPE_MASK_PROGRAMMATIC;
    goto OK1;
  }
  
  if (mask && (!mask->programmatic))
    type = GLC_PROGRAM_TYPE_SIMPLE;

 OK1:
  if ((src->programmatic || src->convolution) &&
      (mask && (mask->programmatic || mask->convolution)))
    return GLC_PROGRAM_TYPE_NOT_SUPPORTED;
  
 OK2:
  return type;
}

void
glc_program_enable (glc_program_type_t type,
                    glc_surface_t *dst,
                    glc_surface_t *src,
                    glc_surface_t *mask,
                    glc_texture_t *src_texture,
                    glc_texture_t *mask_texture)
{
  switch (type) {
  case GLC_PROGRAM_TYPE_SRC_CONVOLUTION:
    glc_program_enable_convolution (dst->proc_address, dst->programs,
                                    src, mask, src_texture, mask_texture,
                                    GLC_PROGRAM_SRC_OPERATION_OFFSET, 0);
    break;
  case GLC_PROGRAM_TYPE_SRC_CONVOLUTION_AND_SOLID_MASK:
    glc_program_enable_convolution (dst->proc_address, dst->programs,
                                    src, mask, src_texture, mask_texture,
                                    GLC_PROGRAM_SRC_OPERATION_OFFSET, 1);
    break;
  case GLC_PROGRAM_TYPE_MASK_CONVOLUTION:
    glc_program_enable_convolution (dst->proc_address, dst->programs,
                                    src, mask, src_texture, mask_texture,
                                    GLC_PROGRAM_MASK_OPERATION_OFFSET, 0);
    break;
  case GLC_PROGRAM_TYPE_MASK_CONVOLUTION_AND_SOLID_SRC:
    glc_program_enable_convolution (dst->proc_address, dst->programs,
                                    src, mask, src_texture, mask_texture,
                                    GLC_PROGRAM_MASK_OPERATION_OFFSET, 2);
    break;
  case GLC_PROGRAM_TYPE_SRC_PROGRAMMATIC:
    glc_program_enable_programmatic (dst, (glc_programmatic_surface_t *) src,
                                     src_texture, mask_texture,
                                     GLC_PROGRAM_SRC_OPERATION_OFFSET);
    break;
  case GLC_PROGRAM_TYPE_MASK_PROGRAMMATIC:
    glc_program_enable_programmatic (dst, (glc_programmatic_surface_t *) mask,
                                     src_texture, mask_texture,
                                     GLC_PROGRAM_MASK_OPERATION_OFFSET);
    break;
  case GLC_PROGRAM_TYPE_SIMPLE:
    glc_program_enable_simple (dst->proc_address, dst->programs,
                               src_texture, mask_texture);
    break;
  case GLC_PROGRAM_TYPE_NONE:
  case GLC_PROGRAM_TYPE_NOT_SUPPORTED:
    break;
  }
}

void
glc_program_disable (glc_surface_t *dst)
{
  if (dst->feature_mask & GLC_FEATURE_ARB_FRAGMENT_PROGRAM_MASK) {
    dst->proc_address->gl_bind_program_arb (GL_FRAGMENT_PROGRAM_ARB, 0);
    glDisable (GL_FRAGMENT_PROGRAM_ARB);
    dst->proc_address->gl_bind_program_arb (GL_VERTEX_PROGRAM_ARB, 0);
    glDisable (GL_VERTEX_PROGRAM_ARB);
  }
}
