/*
 * Copyright  2004 David Reveman, Peter Nilsson
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the names of
 * David Reveman and Peter Nilsson not be used in advertising or
 * publicity pertaining to distribution of the software without
 * specific, written prior permission. David Reveman and Peter Nilsson
 * makes no representations about the suitability of this software for
 * any purpose. It is provided "as is" without express or implied warranty.
 *
 * DAVID REVEMAN AND PETER NILSSON DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL DAVID REVEMAN AND
 * PETER NILSSON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: David Reveman <c99drn@cs.umu.se>
 *          Peter Nilsson <c99pnn@cs.umu.se>
 */

#ifdef HAVE_CONFIG_H
#  include "../config.h"
#endif

#include "glcint.h"

#include <math.h>

static glc_surface_t *
_glc_programmatic_surface_create_similar (void *abstract_templ,
                                          glc_format_name_t format_name,
                                          int width,
                                          int height)
{
  return NULL;
}

static void
_glc_programmatic_surface_destroy (void *abstract_surface)
{
  glc_programmatic_surface_t *surface =
    (glc_programmatic_surface_t *) abstract_surface;
  
  glc_surface_deinit (&surface->base);
  
  free (surface);
}

static glc_bool_t
_glc_programmatic_surface_push_current (void *abstract_surface,
                                        glc_constraint_t constraint)
{
  glc_programmatic_surface_t *surface =
    (glc_programmatic_surface_t *) abstract_surface;
  
  glc_surface_status_add (&surface->base, GLC_STATUS_NOT_SUPPORTED_MASK);
  
  return 0;
}

static void
_glc_programmatic_surface_pop_current (void *abstract_surface) {}

static glc_texture_t *
_glc_programmatic_surface_get_texture (void *abstract_surface)
{
  glc_programmatic_surface_t *surface =
    (glc_programmatic_surface_t *) abstract_surface;
  
  return &surface->texture;
}

static void
_glc_programmatic_surface_realize (void *abstract_surface) {}

static void
_glc_programmatic_surface_show (void *abstract_surface) {}

static const struct glc_surface_backend glc_programmatic_surface_backend = {
  _glc_programmatic_surface_create_similar,
  _glc_programmatic_surface_destroy,
  _glc_programmatic_surface_push_current,
  _glc_programmatic_surface_pop_current,
  _glc_programmatic_surface_get_texture,
  _glc_programmatic_surface_realize,
  _glc_programmatic_surface_show
};

glc_programmatic_surface_t *
_glc_programmatic_surface_create (void)
{
  glc_programmatic_surface_t *surface;

  surface = (glc_programmatic_surface_t *)
    malloc (sizeof (glc_programmatic_surface_t));
  if (surface == NULL)
    return NULL;
  
  glc_surface_init (&surface->base, &glc_programmatic_surface_backend);
  
  surface->base.programmatic = 1;
  surface->base.texture = &surface->texture;
  surface->texture.name = 0;
  surface->texture.target = GL_TEXTURE_2D;
  surface->texture.format = GL_RGBA;
  surface->texture.filter = surface->base.filter;
  surface->texture.texcoord_width =
    surface->texture.texcoord_height = 1.0;
  surface->texture.repeatable = surface->texture.repeat = 1;
  surface->base.repeat = 0;
  
  return surface;
}

void
glc_programmatic_surface_setup (glc_surface_t *abstract_surface,
                                int width,
                                int height)
{
 glc_programmatic_surface_t *surface =
    (glc_programmatic_surface_t *) abstract_surface;
 
 surface->texture.texcoord_width = surface->base.width =
   surface->texture.width = width;
 surface->texture.texcoord_height = surface->base.height =
   surface->texture.height = height;
 surface->base.repeat = 0;
}

glc_surface_t *
glc_solid_create (glc_color_t *color)
{
  glc_programmatic_surface_t *surface;
  
  surface = _glc_programmatic_surface_create ();
  if (!surface)
    return NULL;

  surface->type = GLC_PROGRAMMATIC_SURFACE_SOLID_TYPE;
  surface->attributes.solid.color = *color;
  surface->texture.target = 0;
  
  return &surface->base;
}
slim_hidden_def(glc_solid_create);

/* TODO: support color stops */
glc_surface_t *
glc_linear_gradient_create (glc_colorpoint_t *start,
                            glc_colorpoint_t *end,
                            glc_colorstop_t *stops,
                            int n_stops)
{
  glc_programmatic_surface_t *surface;
  glc_linear_gradient_t *gradient;

  surface = _glc_programmatic_surface_create ();
  if (!surface)
    return NULL;

  surface->type = GLC_PROGRAMMATIC_SURFACE_LINEAR_GRADIENT_TYPE;
  gradient = &surface->attributes.linear_gradient;

  gradient->start = *start;
  gradient->end = *end;    
  
  return &surface->base;
}
slim_hidden_def(glc_linear_gradient_create);

/* TODO: support more than one color stop and adjustable focus point */
glc_surface_t *
glc_radial_gradient_create (glc_colorpoint_t *start,
                            glc_colorstop_t *stops,
                            int n_stops)
{
  glc_programmatic_surface_t *surface;
  glc_radial_gradient_t *gradient;

  surface = _glc_programmatic_surface_create ();
  if (!surface)
    return NULL;

  surface->type = GLC_PROGRAMMATIC_SURFACE_RADIAL_GRADIENT_TYPE;
  gradient = &surface->attributes.radial_gradient;

  gradient->start = *start;

  if (n_stops) {
    gradient->stop = *stops;
  } else {
    gradient->stop.stop = 0;
    gradient->stop.color = gradient->start.color;
  }
  
  return &surface->base;
}
slim_hidden_def(glc_radial_gradient_create);

void
glc_programmatic_surface_set_parameters (glc_gl_proc_address_list_t
                                         *proc_address,
                                         glc_programmatic_surface_t *surface)
{
  switch (surface->type) {
  case GLC_PROGRAMMATIC_SURFACE_SOLID_TYPE: {
    glc_solid_t *solid = &surface->attributes.solid;
    
    glColor4us (solid->color.red, solid->color.green, solid->color.blue,
                solid->color.alpha);
  } break;
  case GLC_PROGRAMMATIC_SURFACE_LINEAR_GRADIENT_TYPE: {
    glc_linear_gradient_t *gradient = &surface->attributes.linear_gradient;
    glc_point_t p1, p2;
    double length, angle, start, end;

    p1.x = FIXED_TO_DOUBLE (gradient->start.point.x);
    p1.y = surface->base.height - FIXED_TO_DOUBLE (gradient->start.point.y);
    
    p2.x = FIXED_TO_DOUBLE (gradient->end.point.x);
    p2.y = surface->base.height - FIXED_TO_DOUBLE (gradient->end.point.y);

    length = sqrt ((p2.x - p1.x) * (p2.x - p1.x) +
                   (p2.y - p1.y) * (p2.y - p1.y));
  
    angle = -atan2 (p2.y - p1.y, p2.x - p1.x);

    start = cos (angle) * p1.x;
    start += -sin (angle) * p1.y;

    end = cos (angle) * p2.x;
    end += -sin (angle) * p2.y;

    proc_address->gl_program_local_param_4d_arb
      (GL_FRAGMENT_PROGRAM_ARB, 0,
       FIXED_TO_DOUBLE (gradient->start.color.red),
       FIXED_TO_DOUBLE (gradient->start.color.green),
       FIXED_TO_DOUBLE (gradient->start.color.blue),
       FIXED_TO_DOUBLE (gradient->start.color.alpha));
    proc_address->gl_program_local_param_4d_arb
      (GL_FRAGMENT_PROGRAM_ARB, 1,
       FIXED_TO_DOUBLE (gradient->end.color.red),
       FIXED_TO_DOUBLE (gradient->end.color.green),
       FIXED_TO_DOUBLE (gradient->end.color.blue),
       FIXED_TO_DOUBLE (gradient->end.color.alpha));
    proc_address->gl_program_local_param_4d_arb
      (GL_FRAGMENT_PROGRAM_ARB, 2,
       start,
       end,
       1.0 / length,
       0.0);
    proc_address->gl_program_local_param_4d_arb
      (GL_FRAGMENT_PROGRAM_ARB, 3,
       cos (angle),
       -sin (angle),
       0.0, 0.0);
  } break;
  case GLC_PROGRAMMATIC_SURFACE_RADIAL_GRADIENT_TYPE: {
    glc_radial_gradient_t *gradient = &surface->attributes.radial_gradient;
    double length = FIXED_TO_DOUBLE (gradient->stop.stop);

    proc_address->gl_program_local_param_4d_arb
      (GL_FRAGMENT_PROGRAM_ARB, 0,
       FIXED_TO_DOUBLE (gradient->start.color.red),
       FIXED_TO_DOUBLE (gradient->start.color.green),
       FIXED_TO_DOUBLE (gradient->start.color.blue),
       FIXED_TO_DOUBLE (gradient->start.color.alpha));
    proc_address->gl_program_local_param_4d_arb
      (GL_FRAGMENT_PROGRAM_ARB, 1,
       FIXED_TO_DOUBLE (gradient->stop.color.red),
       FIXED_TO_DOUBLE (gradient->stop.color.green),
       FIXED_TO_DOUBLE (gradient->stop.color.blue),
       FIXED_TO_DOUBLE (gradient->stop.color.alpha));
    proc_address->gl_program_local_param_4d_arb
      (GL_FRAGMENT_PROGRAM_ARB, 2,
       FIXED_TO_DOUBLE (gradient->start.point.x),
       surface->base.height - FIXED_TO_DOUBLE (gradient->start.point.y),
       (length)? 1.0 / (length * length): 1000000.0,
       0.0);
  } break;
  default:
    break;
  }
}
