/*
 * Copyright  2004 David Reveman, Peter Nilsson
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the names of
 * David Reveman and Peter Nilsson not be used in advertising or
 * publicity pertaining to distribution of the software without
 * specific, written prior permission. David Reveman and Peter Nilsson
 * makes no representations about the suitability of this software for
 * any purpose. It is provided "as is" without express or implied warranty.
 *
 * DAVID REVEMAN AND PETER NILSSON DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL DAVID REVEMAN AND
 * PETER NILSSON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: David Reveman <c99drn@cs.umu.se>
 *          Peter Nilsson <c99pnn@cs.umu.se>
 */

#ifdef HAVE_CONFIG_H
#  include "../config.h"
#endif

#include "glcint.h"

#include <stdlib.h>
#include <string.h>

void
glc_surface_init (glc_surface_t *surface,
                  const glc_surface_backend_t *backend)
{
  surface->backend = backend;

  surface->format = NULL;
  surface->texture = NULL;
  surface->feature_mask = 0;
  surface->repeat = 0;
  surface->transform = NULL;
  surface->transforms = NULL;
  surface->n_transforms = 0;
  surface->anti_aliasing = 0;
  surface->filter = GLC_FILTER_NEAREST;
  surface->polyedge = GLC_POLYEDGE_SMOOTH;
  surface->status_mask = 0;
  surface->implicit_mask = 0;
  surface->dirty = 0;
  surface->requires_flipping = 1;
  surface->convolution = NULL;
  surface->clear_exterior = 0;
  surface->programmatic = 0;
}

void
glc_surface_deinit (glc_surface_t *surface)
{
  if (surface->transforms)
    free (surface->transforms);

  if (surface->convolution)
    free (surface->convolution);
}

glc_surface_t *
glc_surface_create_similar (glc_surface_t *templ,
                            glc_format_name_t format_name,
                            int width,
                            int height)
{
  glc_surface_t *surface =
    templ->backend->create_similar (templ, format_name, width, height);

  if (surface)
    surface->polyedge = templ->polyedge;
  
  return surface;
}

void
glc_surface_destroy (glc_surface_t *surface)
{
  surface->backend->destroy (surface);
}

glc_bool_t
glc_surface_push_current (glc_surface_t *surface,
                          glc_constraint_t constraint)
{
  if (!surface->backend->push_current (surface, constraint)) {
    glc_surface_status_add (surface, GLC_STATUS_NOT_SUPPORTED_MASK);
    return 0;
  }

  return 1;
}

void
glc_surface_pop_current (glc_surface_t *surface)
{
  surface->backend->pop_current (surface);
}

void
glc_surface_enable_program (glc_program_type_t type,
                            glc_surface_t *surface,
                            glc_surface_t *src,
                            glc_surface_t *mask,
                            glc_texture_t *src_texture,
                            glc_texture_t *mask_texture)
{
  return
    glc_program_enable (type, surface, src, mask, src_texture, mask_texture);
}

void
glc_surface_disable_program (glc_surface_t *surface)
{
  glc_program_disable (surface);
}

void
glc_surface_push_transform (glc_surface_t *surface)
{
  static const glc_matrix_t identity = {
    {
      { 1.0, 0.0, 0.0 },
      { 0.0, 1.0, 0.0 },
      { 0.0, 0.0, 1.0 }
    }
  };
  
  surface->n_transforms++;
  
  surface->transforms =
    realloc (surface->transforms,
             surface->n_transforms * sizeof (glc_matrix_t));

  if (!surface->transforms)
    return;
  
  if (surface->n_transforms > 1) {
    surface->transforms[surface->n_transforms - 1] =
      surface->transforms[surface->n_transforms - 2];
  } else
    surface->transforms[surface->n_transforms - 1] = identity;
  
  surface->transform = &surface->transforms[surface->n_transforms - 1];
}

void
glc_surface_pop_transform (glc_surface_t *surface)
{
  if (surface->n_transforms < 1)
    return;
  
  surface->n_transforms--;
  
  surface->transforms =
    realloc (surface->transforms,
             surface->n_transforms * sizeof (glc_matrix_t));

  if (surface->n_transforms)
    surface->transform = &surface->transforms[surface->n_transforms - 1];
  else
    surface->transform = NULL;
}

void
glc_surface_bounds (glc_surface_t *surface,
                    glc_region_box_t *box)
{
  box->y1 = MINSHORT;
  box->y2 = MAXSHORT;
  box->x1 = MINSHORT;
  box->x2 = MAXSHORT;
  
  if (surface->programmatic || surface->repeat || surface->transform)
    return;
    
  box->x1 = 0;
  box->y1 = 0;
  box->x2 = surface->width;
  box->y2 = surface->height;
}

void
glc_surface_set_transform (glc_surface_t *surface,
                           glc_transform_t *transform)
{
  static const glc_transform_t identity = {
    {
      { FIXED1, 0x00000, 0x00000 },
      { 0x00000, FIXED1, 0x00000 },
      { 0x00000, 0x00000, FIXED1 }
    }
  };

  if (surface->programmatic)
    return;

  if (transform && memcmp (transform, &identity,
                           sizeof (glc_transform_t)) == 0)
    transform = NULL;
  
  if (transform) {
    if (!surface->transform) {
      glc_surface_push_transform (surface);
      if (!surface->transform)
		return;
	}

    surface->transform->m[0][0] = FIXED_TO_DOUBLE (transform->matrix[0][0]);
    surface->transform->m[1][0] = FIXED_TO_DOUBLE (transform->matrix[0][1]);
    surface->transform->m[2][0] = FIXED_TO_DOUBLE (transform->matrix[0][2]);

    surface->transform->m[0][1] = FIXED_TO_DOUBLE (transform->matrix[1][0]);
    surface->transform->m[1][1] = FIXED_TO_DOUBLE (transform->matrix[1][1]);
    surface->transform->m[2][1] = FIXED_TO_DOUBLE (transform->matrix[1][2]);

    if (glc_matrix_invert (surface->transform)) {
      glc_surface_pop_transform (surface);
      glc_surface_status_add (surface, GLC_STATUS_INVALID_MATRIX_MASK);
    }
  } else
    glc_surface_pop_transform (surface);
}
slim_hidden_def(glc_surface_set_transform);

void
glc_surface_set_convolution (glc_surface_t *surface,
                             glc_convolution_t *convolution)
{
  if (surface->programmatic)
    return;
  
  if (convolution &&
      convolution->matrix[0][0] == 0x00000 &&
      convolution->matrix[0][1] == 0x00000 &&
      convolution->matrix[0][2] == 0x00000 &&
      convolution->matrix[1][0] == 0x00000 &&
      convolution->matrix[1][2] == 0x00000 &&
      convolution->matrix[2][0] == 0x00000 &&
      convolution->matrix[2][1] == 0x00000 &&
      convolution->matrix[2][2] == 0x00000)
    convolution = NULL;
  
  if (convolution) {
    int row, col;
    
    if (!surface->convolution) {
      surface->convolution = malloc (sizeof (glc_matrix_t));
      if (!surface->convolution)
		return;
	}

    for (row = 0; row < 3; row++)
      for (col = 0; col < 3; col++)
        surface->convolution->m[row][col] =
          FIXED_TO_DOUBLE (convolution->matrix[row][col]);
    
    if (glc_matrix_normalize (surface->convolution)) {
      free (surface->convolution);
      surface->convolution = NULL;
      glc_surface_status_add (surface, GLC_STATUS_INVALID_MATRIX_MASK);
    }
  } else {
    if (surface->convolution) {
      free (surface->convolution);
      surface->convolution = NULL;
    }
  }
}
slim_hidden_def(glc_surface_set_convolution);

void
glc_surface_set_repeat (glc_surface_t *surface,
                        glc_bool_t repeat)
{
  if (surface->programmatic)
    return;
  
  surface->repeat = repeat;
}
slim_hidden_def(glc_surface_set_repeat);

void
glc_surface_set_filter (glc_surface_t *surface,
                        glc_filter_t filter)
{
  if (surface->programmatic)
    return;
  
  surface->filter = filter;
}
slim_hidden_def(glc_surface_set_filter);

void
glc_surface_set_polyedge (glc_surface_t *surface,
                          glc_polyedge_t polyedge)
{
  surface->polyedge = polyedge;
}
slim_hidden_def(glc_surface_set_polyedge);

int
glc_surface_get_width (glc_surface_t *surface)
{
  if (surface->programmatic)
    return 1;
  
  return surface->width;
}
slim_hidden_def(glc_surface_get_width);

int
glc_surface_get_height (glc_surface_t *surface)
{
  if (surface->programmatic)
    return 1;
  
  return surface->height;
}
slim_hidden_def(glc_surface_get_height);

glc_texture_t *
glc_surface_get_texture (glc_surface_t *surface)
{
  return surface->backend->get_texture (surface);
}

void
glc_surface_realize (glc_surface_t *surface)
{
  surface->backend->realize (surface);
}
slim_hidden_def(glc_surface_realize);

static void
_glc_set_raster_pos (int x, int y)
{
  glPushAttrib (GL_TRANSFORM_BIT | GL_VIEWPORT_BIT);
  glMatrixMode (GL_PROJECTION);
  glPushMatrix ();
  glLoadIdentity ();
  glMatrixMode (GL_MODELVIEW);
  glPushMatrix ();
  glLoadIdentity ();
  glDepthRange (0, 1);
  glViewport (-1, -1, 2, 2);
  
  glRasterPos2d (0, 0);
  glBitmap (0, 0, 1, 1, x, y, NULL);
  
  glPopMatrix ();
  glMatrixMode (GL_PROJECTION);
  glPopMatrix ();
  glPopAttrib ();
}

void
glc_surface_show (glc_surface_t *surface,
                  int x,
                  int y,
                  unsigned int width,
                  unsigned int height)
{
  if (surface->programmatic)
    return;
  
  if (surface->format->doublebuffer &&
      x <= 0 && y <= 0 &&
      (x + (int) width) >= surface->width &&
      (y + (int) height) >= surface->height) {
    surface->backend->show (surface);
  } else if (width > 0 && height > 0) {
    if (x < 0) x = 0;
    if (y < 0) y = 0;
    if (glc_surface_push_current (surface, GLC_CN_SURFACE_DRAWABLE_CURRENT)) {
      if (surface->format->doublebuffer) {
        glReadBuffer (GL_BACK);
        glDrawBuffer (GL_FRONT);
        glDisable (GL_SCISSOR_TEST);
        glDisable (GL_DITHER);
        glc_set_operator (GLC_OPERATOR_SRC);

        _glc_set_raster_pos (x, surface->height - (y + height));
        glCopyPixels (x, surface->height - (y + height),
                      width, height, GL_COLOR);
        
        glDrawBuffer (GL_BACK);
      }
      glFlush ();
    }
    glc_surface_pop_current (surface);
  }
}
slim_hidden_def(glc_surface_show);

void
glc_surface_dirty (glc_surface_t *surface,
                   glc_region_box_t *region)
{
  if (!region) {
    surface->dirty_region.x1 = surface->dirty_region.y1 = 0;
    surface->dirty_region.x2 = surface->width;
    surface->dirty_region.y2 = surface->height;
  } else {
    if (!surface->dirty) {
      surface->dirty_region = *region;
    } else
      glc_union_region (region,
                        &surface->dirty_region,
                        &surface->dirty_region);
  }
  
  surface->dirty = 1;
  
  glFlush ();
}

void
glc_surface_status_add (glc_surface_t *surface, int flags)
{
  surface->status_mask |= flags;
}

glc_status_t
glc_surface_get_status (glc_surface_t *surface)
{
  return glc_status_pop_from_mask (&surface->status_mask);
}
slim_hidden_def(glc_surface_get_status);

void
glc_surface_setup_environment (glc_surface_t *surface)
{
  glPixelStorei (GL_PACK_ALIGNMENT, 4);
  glPixelStorei (GL_UNPACK_ALIGNMENT, 4);
  glViewport (0, 0, surface->width, surface->height);
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  glOrtho (0.0, surface->width, 0.0, surface->height, -1.0, 1.0);
  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();
  glScalef (1, -1, 1);
  glTranslatef (0, -surface->height, 0);
  glDisable (GL_DEPTH_TEST); 
  glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
  glDisable (GL_CULL_FACE);
  glDepthMask (GL_FALSE);
  glEnable (GL_SCISSOR_TEST);
  glScissor (0, 0, surface->width, surface->height);
  glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
  glDisable (GL_POLYGON_SMOOTH);
  glShadeModel (GL_FLAT);
  glColorMask (1, 1, 1, 1);
}

void
glc_surface_read_pixels (glc_surface_t *surface,
                         int x,
                         int y,
                         unsigned int width,
                         unsigned int height,
                         char *pixels)
{
  unsigned char *pixel_buf;
  int rowstride, area_rowstride, bytes_per_pixel;
  unsigned int i;
  GLenum format, type;

  if (surface->programmatic)
    return;

  if (x < 0 || x > (int) surface->width - (int) width ||
      y < 0 || y > (int) surface->height - (int) height) {
    glc_surface_status_add (surface, GLC_STATUS_BAD_COORDINATE_MASK);
    return;
  }

  bytes_per_pixel = (surface->format->bpp / 8);
  format = glc_get_gl_format_from_bpp (surface->format->bpp);
  type = glc_get_gl_data_type_from_bpp (surface->format->bpp);

  /* We currently read the whole image to a temporary buffer and then
     copy the part we want, not very efficient. We only want to read the
     area requested. I think it can be fixed with glPixelStore parameters. */

  if (glc_surface_push_current (surface, GLC_CN_SURFACE_DRAWABLE_CURRENT)) {
    rowstride = surface->width * bytes_per_pixel;
    rowstride += (rowstride % 4)? (4 - (rowstride % 4)): 0;
    pixel_buf = (unsigned char *) malloc (surface->height * rowstride);
  
    glReadPixels (0, 0, surface->width, surface->height,
                  format, type, pixel_buf);
  } else {
    glc_texture_t *texture;

    texture = glc_surface_get_texture (surface);
    glc_texture_bind (texture);

    rowstride = texture->width * bytes_per_pixel;
    rowstride += (rowstride % 4)? (4 - (rowstride % 4)): 0;
    pixel_buf = (unsigned char *) malloc (texture->height * rowstride);

    glPixelStorei (GL_UNPACK_ROW_LENGTH, 0);
    glPixelStorei (GL_UNPACK_SKIP_ROWS, 0);
    glPixelStorei (GL_UNPACK_SKIP_PIXELS, 0);
    
    glGetTexImage (texture->target, 0,
                   format,
                   type,
                   pixel_buf);

    glc_texture_unbind (texture);
  }

  glc_surface_pop_current (surface);

  area_rowstride = width * bytes_per_pixel;
  area_rowstride += (area_rowstride % 4)? (4 - (area_rowstride % 4)): 0;

  for (i = 0; i < height; i++) {
    memcpy (&pixels[(height - i - 1) * area_rowstride],
            &pixel_buf[(surface->height - y - height + i) * rowstride + x],
            area_rowstride);
  }
  
  free (pixel_buf);
}
slim_hidden_def(glc_surface_read_pixels);

void
glc_surface_draw_pixels (glc_surface_t *surface,
                         int x,
                         int y,
                         unsigned int width,
                         unsigned int height,
                         char *pixels)
{
  unsigned char *pixel_buf;
  GLenum format, type;
  int bytes_per_pixel;

  if (surface->programmatic)
    return;

  if (x < 0 || x > (surface->width - (int) width) ||
      y < 0 || y > (surface->height - (int) height)) {
    glc_surface_status_add (surface, GLC_STATUS_BAD_COORDINATE_MASK);
    return;
  }

  bytes_per_pixel = (surface->format->bpp / 8);
  format = glc_get_gl_format_from_bpp (surface->format->bpp);
  type = glc_get_gl_data_type_from_bpp (surface->format->bpp);

  if (glc_surface_push_current (surface, GLC_CN_SURFACE_DRAWABLE_CURRENT)) {
    glc_region_box_t bounds;

    bounds.x1 = x;
    bounds.x2 = x + width;
    bounds.y1 = y;
    bounds.y2 = y + height;

    glDisable (GL_SCISSOR_TEST);
    glDisable (GL_DITHER);
    glc_set_operator (GLC_OPERATOR_SRC);

    if (surface->requires_flipping)
      glPixelZoom (1.0, -1.0);
    
    _glc_set_raster_pos (x, surface->height - y);
    glDrawPixels (width, height, format, type, pixels);

    glc_surface_dirty (surface, &bounds);
  } else {
    int i, rowstride;
    glc_texture_t *texture;
    
    rowstride = width * bytes_per_pixel;
    rowstride += (rowstride % 4)? (4 - (rowstride % 4)): 0;
    pixel_buf = (unsigned char *) malloc (rowstride * height);
    
    /* TODO: This is very ugly and I like to remove it as soon as possible.
       Needs some changes to texture handling, as images will become upside
       down without it. */
    for (i = 0; i < (int) height; i++)
      memcpy (&pixel_buf[i * rowstride],
              &pixels[(height - i - 1) * rowstride], rowstride);

    texture = glc_surface_get_texture (surface);
    glc_texture_bind (texture);
    
    glPixelStorei (GL_PACK_ROW_LENGTH, 0);
    glPixelStorei (GL_PACK_SKIP_ROWS, 0);
    glPixelStorei (GL_PACK_SKIP_PIXELS, 0);
    
    glTexSubImage2D (texture->target, 0,
                     x, surface->height - y - height,
                     width, height,
                     format, type,
                     pixel_buf);
    glFlush ();

    glc_texture_unbind (texture);
    
    free (pixel_buf);
  }

  glc_surface_pop_current (surface);
}
slim_hidden_def(glc_surface_draw_pixels);

void
glc_surface_get_gl_texture (glc_surface_t *surface,
                            unsigned int *name,
                            unsigned int *target,
                            double *texcoord_width,
                            double *texcoord_height,
                            glc_bool_t *repeatable)
{
  glc_texture_t *texture = glc_surface_get_texture (surface);
  
  if (!texture)
    return;

  if (!surface->programmatic) {
    glc_texture_bind (texture);
    glc_texture_ensure_filter (texture, surface->filter);
    glc_texture_ensure_repeat (texture,
                               (surface->repeat && texture->repeatable));
    glc_texture_unbind (texture);
  }

  if (name)
    *name = texture->name;

  if (target)
    *target = texture->target;

  if (texcoord_width)
    *texcoord_width = texture->texcoord_width;

  if (texcoord_height)
    *texcoord_height = texture->texcoord_height;

  if (repeatable)
    *repeatable = texture->repeatable;
}
slim_hidden_def(glc_surface_get_gl_texture);

void
glc_surface_gl_begin (glc_surface_t *surface)
{
  if (!glc_surface_push_current (surface, GLC_CN_SURFACE_DRAWABLE_CURRENT))
    glc_surface_status_add (surface, GLC_STATUS_NOT_SUPPORTED_MASK);
}
slim_hidden_def(glc_surface_gl_begin);

void
glc_surface_gl_end (glc_surface_t *surface)
{
  glc_region_box_t region;

  region.x1 = region.y1 = 0;
  region.x2 = surface->width;
  region.y2 = surface->height;
  
  glc_surface_dirty (surface, &region);
  
  glc_surface_pop_current (surface);
}
slim_hidden_def(glc_surface_gl_end);

unsigned long
glc_surface_features (glc_surface_t *surface)
{
  return surface->feature_mask;
}
slim_hidden_def(glc_surface_features);
