/*
 * Copyright  2004 David Reveman, Peter Nilsson
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the names of
 * David Reveman and Peter Nilsson not be used in advertising or
 * publicity pertaining to distribution of the software without
 * specific, written prior permission. David Reveman and Peter Nilsson
 * makes no representations about the suitability of this software for
 * any purpose. It is provided "as is" without express or implied warranty.
 *
 * DAVID REVEMAN AND PETER NILSSON DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL DAVID REVEMAN AND
 * PETER NILSSON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: David Reveman <c99drn@cs.umu.se>
 *          Peter Nilsson <c99pnn@cs.umu.se>
 */

#ifdef HAVE_CONFIG_H
#  include "../config.h"
#endif

#include "glcint.h"

static glc_bool_t
_glc_texture_is_power_of_two (int value)
{
  int x = 1;
  
  while (x < value)
    x <<= 1;

  return (x == value);
}

static void
_glc_texture_find_best_target (int width, int height,
                               long int target_mask,
                               unsigned int *target)
{
  *target = GL_TEXTURE_2D;

  if ((!(target_mask & GLC_TEXTURE_TARGET_2D_MASK)) ||
      (!_glc_texture_is_power_of_two (width)) ||
      (!_glc_texture_is_power_of_two (height))) {
    if (target_mask & GLC_TEXTURE_TARGET_RECTANGLE_MASK)
      *target = GL_TEXTURE_RECTANGLE_EXT;
  }
}

static void
_glc_texture_power_of_two_size (int *value)
{
  int x = 1;
  
  while (x < *value)
    x <<= 1;

  *value = x;
}

glc_texture_t *
glc_texture_generate (int width,
                      int height,
                      unsigned int texture_format,
                      long int target_mask)
{
  glc_texture_t *texture;

  texture = (glc_texture_t *) malloc (sizeof (glc_texture_t));
  if (texture == NULL)
    return NULL;

  texture->filter = -1;
  texture->repeat = -1;
  texture->width = width;
  texture->height = height;
  texture->format = texture_format;
  texture->allocated = 0;

  if (!(target_mask & GLC_TEXTURE_TARGET_NPOT_MASK)) {
    _glc_texture_find_best_target (width, height,
                                   target_mask, &texture->target);
    
    if (texture->target == GL_TEXTURE_2D) {
      _glc_texture_power_of_two_size (&texture->width);
      _glc_texture_power_of_two_size (&texture->height);
    }
  } else
    texture->target = GL_TEXTURE_2D;
  
  glGenTextures (1, (GLuint *) &texture->name);
  
  if (texture->target == GL_TEXTURE_2D &&
      texture->width == width && texture->height == height) {
    texture->repeatable = 1;
    texture->texcoord_width = texture->texcoord_height = 1.0;
  } else {
    texture->repeatable = 0;
    if (texture->target == GL_TEXTURE_2D) {
      texture->texcoord_width = (double) width / (double) texture->width;
      texture->texcoord_height = (double) height / (double) texture->height;
    } else {
      texture->texcoord_width = texture->width;
      texture->texcoord_height = texture->height;
    }
  }
  
  return texture;
}

void
glc_texture_allocate (glc_texture_t *texture)
{
  if (texture->allocated)
    return;

  glc_texture_bind (texture);

  glTexImage2D (texture->target, 0, GL_RGBA,
                texture->width, texture->height,
                0, texture->format, GL_UNSIGNED_BYTE, NULL);

  glc_texture_unbind (texture);

  texture->allocated = 1;
}

void 
glc_texture_destroy (glc_texture_t *texture)
{ 
  glDeleteTextures (1, (GLuint *) &texture->name);
  free (texture);
}

void
glc_texture_ensure_filter (glc_texture_t *texture,
                           glc_filter_t filter)
{
  if (!texture->target)
    return;
    
  if (texture->filter != filter) {
    switch (filter) {
    case GLC_FILTER_FAST:
    case GLC_FILTER_NEAREST:
      glTexParameteri (texture->target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
      glTexParameteri (texture->target, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
      break;
    case GLC_FILTER_GOOD:
    case GLC_FILTER_BEST:
    case GLC_FILTER_BILINEAR:
      glTexParameteri (texture->target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
      glTexParameteri (texture->target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
      break;
    }
    texture->filter = filter;
  }
}

void
glc_texture_ensure_repeat (glc_texture_t *texture,
                           glc_bool_t repeat)
{
  if (!texture->target)
    return;
  
  if (texture->repeat != repeat) {
    if (repeat) {
      glTexParameteri (texture->target, GL_TEXTURE_WRAP_S, GL_REPEAT);
      glTexParameteri (texture->target, GL_TEXTURE_WRAP_T, GL_REPEAT);
    } else {
      glTexParameteri (texture->target, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
      glTexParameteri (texture->target, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    }
    texture->repeat = repeat;
  }
}

void
glc_texture_bind (glc_texture_t *texture)
{
  
  glDisable (GL_TEXTURE_RECTANGLE_EXT);
  glDisable (GL_TEXTURE_2D);

  if (!texture->target)
    return;
  
  glEnable (texture->target);
  glBindTexture (texture->target, texture->name);
}

void
glc_texture_unbind (glc_texture_t *texture)
{
  glBindTexture (texture->target, 0);
  glDisable (texture->target);
}

void
glc_texture_copy_surface (glc_texture_t *texture,
                          glc_surface_t *surface,
                          glc_region_box_t *region)
{
  glc_surface_push_current (surface, GLC_CN_SURFACE_DRAWABLE_CURRENT);
  
  glc_texture_bind (texture);

  if (region->x1 < 0) region->x1 = 0;
  if (region->y1 < 0) region->y1 = 0;
  if (region->x2 > surface->width) region->x2 = surface->width;
  if (region->y2 > surface->height) region->y2 = surface->height;

  glCopyTexSubImage2D (texture->target, 0,
                       region->x1, region->y1,
                       region->x1, region->y1,
                       region->x2 - region->x1,
                       region->y2 - region->y1);
  
  glFlush ();

  glc_texture_unbind (texture);
  glc_surface_pop_current (surface);
}
