/*
 * Copyright  2004 David Reveman, Peter Nilsson
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the names of
 * David Reveman and Peter Nilsson not be used in advertising or
 * publicity pertaining to distribution of the software without
 * specific, written prior permission. David Reveman and Peter Nilsson
 * makes no representations about the suitability of this software for
 * any purpose. It is provided "as is" without express or implied warranty.
 *
 * DAVID REVEMAN AND PETER NILSSON DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL DAVID REVEMAN AND
 * PETER NILSSON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: David Reveman <c99drn@cs.umu.se>
 *          Peter Nilsson <c99pnn@cs.umu.se>
 */

#ifdef HAVE_CONFIG_H
#  include "../config.h"
#endif

#include "glcint.h"

/* whether 't' is a well defined not obviously empty trapezoid */
#define TRAPEZOID_VALID(t) ((t)->left.p1.y != (t)->left.p2.y && \
			     (t)->right.p1.y != (t)->right.p2.y && \
			     (int) ((t)->bottom - (t)->top) > 0)

/* whether 't' is a well defined not obviously empty color trapezoid */
#define COLORTRAPEZOID_VALID(t) ((t)->top.left <= (t)->top.right && \
			     (t)->bottom.left <= (t)->bottom.right && \
			     (int) ((t)->bottom.y - (t)->top.y) >= 0)

static glc_fixed16_16_t
glc_line_fixed_x (const glc_line_fixed_t *l,
                  glc_fixed16_16_t y,
                  int ceil)
{
  glc_fixed_32_32 ex = (glc_fixed_32_32) (y - l->p1.y) * (l->p2.x - l->p1.x);
  glc_fixed16_16_t dy = l->p2.y - l->p1.y;
  
  if (ceil)
    ex += (dy - 1);
  
  return l->p1.x + (glc_fixed16_16_t) (ex / dy);
}

static void
glc_trapezoid_bounds (int n_traps,
                      const glc_trapezoid_t *traps,
                      glc_region_box_t *box)
{
  box->y1 = MAXSHORT;
  box->y2 = MINSHORT;
  box->x1 = MAXSHORT;
  box->x2 = MINSHORT;
  
  for (; n_traps; n_traps--, traps++) {
    int16_t x1, y1, x2, y2;

    if (!TRAPEZOID_VALID (traps))
      continue;
    
    y1 = FIXED_TO_INT (traps->top);
    if (y1 < box->y1)
      box->y1 = y1;

    y2 = FIXED_TO_INT (FIXED_CEIL (traps->bottom));
    if (y2 > box->y2)
      box->y2 = y2;

    x1 = FIXED_TO_INT (MIN (glc_line_fixed_x (&traps->left, traps->top, 0),
                            glc_line_fixed_x (&traps->left,
                                              traps->bottom, 0)));
    if (x1 < box->x1)
      box->x1 = x1;

    x2 = FIXED_TO_INT (FIXED_CEIL
                       (MAX (glc_line_fixed_x (&traps->right, traps->top, 1),
                             glc_line_fixed_x (&traps->right,
                                               traps->bottom, 1))));
    if (x2 > box->x2)
      box->x2 = x2;
  }
}

void
glc_composite_trapezoids (glc_operator_t op,
                          glc_surface_t *src,
                          glc_surface_t *dst,
                          int x_src,
                          int y_src,
                          const glc_trapezoid_t *traps,
                          int n_traps)
{
  glc_surface_t *mask;
  glc_region_box_t trap_bounds, src_bounds, dst_bounds, bounds;
  double x_draw_offset, y_draw_offset;
  int x_dst, y_dst;

  if (n_traps == 0)
	return;

  x_dst = traps[0].left.p1.x >> 16;
  y_dst = traps[0].left.p1.y >> 16;

  glc_trapezoid_bounds (n_traps, traps, &trap_bounds);
  glc_surface_bounds (src, &src_bounds);
  glc_surface_bounds (dst, &dst_bounds);

  src_bounds.x1 += (x_dst - x_src);
  src_bounds.y1 += (y_dst - y_src);
  src_bounds.x2 += (x_dst - x_src);
  src_bounds.y2 += (y_dst - y_src);

  glc_intersect_region (&src_bounds, &trap_bounds, &bounds);
  glc_intersect_region (&dst_bounds, &bounds, &bounds);

  if ((bounds.x2 - bounds.x1) <= 0 || (bounds.y2 - bounds.y1) <= 0)
    return;
  
  x_draw_offset = (double) -bounds.x1;
  y_draw_offset = (double) -bounds.y1;
  
  mask = glc_surface_create_similar (dst, GLC_STANDARD_A8,
                                     bounds.x2 - bounds.x1,
                                     bounds.y2 - bounds.y1);
  
  if (!mask) {
    glc_surface_status_add (dst, GLC_STATUS_NOT_SUPPORTED_MASK);
    return;
  }

  mask->implicit_mask = 1;

  if (!glc_surface_push_current (mask, GLC_CN_SURFACE_DRAWABLE_CURRENT)) {
    glc_surface_pop_current (mask);
    return;
  }
  
  glClearColor (0.0, 0.0, 0.0, 0.0);
  glClear (GL_COLOR_BUFFER_BIT);
  glColor4f (1.0, 1.0, 1.0, 1.0);

  glc_set_operator (GLC_OPERATOR_SRC);

  if (mask->polyedge == GLC_POLYEDGE_SMOOTH) {
    if (!mask->anti_aliasing) {
      glEnable (GL_POLYGON_SMOOTH);
      glEnable (GL_BLEND);
      glHint (GL_POLYGON_SMOOTH, GL_NICEST);
      glBlendFunc (GL_SRC_ALPHA_SATURATE, GL_ONE);
    }
  }

  glBegin (GL_QUADS);

  for (; n_traps; n_traps--, traps++) {
    double top, bottom;
    
    if (!TRAPEZOID_VALID (traps))
      continue;
    
    top =  y_draw_offset + FIXED_TO_DOUBLE (traps->top);
    bottom =  y_draw_offset + FIXED_TO_DOUBLE (traps->bottom);

    glVertex2d (x_draw_offset +
                FIXED_TO_DOUBLE (glc_line_fixed_x
                                 (&traps->left, traps->top, 0)), top);
    glVertex2d (x_draw_offset +
                FIXED_TO_DOUBLE (glc_line_fixed_x
                                 (&traps->right, traps->top, 1)), top);
    glVertex2d (x_draw_offset +
                FIXED_TO_DOUBLE (glc_line_fixed_x
                                 (&traps->right, traps->bottom, 1)), bottom);
    glVertex2d (x_draw_offset +
                FIXED_TO_DOUBLE (glc_line_fixed_x
                                 (&traps->left, traps->bottom, 0)), bottom);
  }

  glEnd ();

  if (mask->polyedge == GLC_POLYEDGE_SMOOTH) {
    if (!mask->anti_aliasing) {
      glDisable (GL_POLYGON_SMOOTH);
      glDisable (GL_BLEND);
    }
  }

  glc_surface_dirty (mask, NULL);
  glc_surface_pop_current (mask);

  glc_composite (op,
                 src,
                 mask,
                 dst,
                 x_src + trap_bounds.x1 - x_dst,
                 y_src + trap_bounds.y1 - y_dst,
                 0, 0,
                 bounds.x1, bounds.y1,
                 mask->width, mask->height);              
  
  glc_surface_destroy (mask);
}
slim_hidden_def(glc_composite_trapezoids);

static void
glc_color_trapezoid_bounds (int n_color_traps,
                            const glc_color_trapezoid_t *color_traps,
                            glc_region_box_t *box)
{
  box->y1 = MAXSHORT;
  box->y2 = MINSHORT;
  box->x1 = MAXSHORT;
  box->x2 = MINSHORT;
  
  for (; n_color_traps; n_color_traps--, color_traps++) {
    int16_t x1, y1, x2, y2;

    if (!COLORTRAPEZOID_VALID (color_traps))
      continue;
    
    x1 = MIN (FIXED_TO_INT (color_traps->top.left),
              FIXED_TO_INT (color_traps->bottom.left));
    if (x1 < box->x1)
      box->x1 = x1;
    
    x2 = MAX (FIXED_TO_INT (FIXED_CEIL (color_traps->top.right)),
              FIXED_TO_INT (FIXED_CEIL (color_traps->bottom.right)));
    if (x2 > box->x2)
      box->x2 = x2;
    
    y1 = FIXED_TO_INT (color_traps->top.y);
    if (y1 < box->y1)
      box->y1 = y1;
    
    y2 = FIXED_TO_INT (FIXED_CEIL (color_traps->bottom.y));
    if (y2 > box->y2)
      box->y2 = y2;
  }
}

void
glc_color_trapezoids (glc_operator_t op,
                      glc_surface_t *dst,
                      const glc_color_trapezoid_t *color_traps,
                      int n_color_traps)
{
  glc_region_box_t bounds;
  int index;
  glc_bool_t shade = 0;
  
  glc_color_trapezoid_bounds (n_color_traps, color_traps, &bounds);
  if (bounds.y1 >= bounds.y2 || bounds.x1 >= bounds.x2 ||
      bounds.x1 > dst->width || bounds.y1 > dst->height ||
      bounds.x2 < 0 || bounds.y2 < 0)
    return;
  
  if (!glc_surface_push_current (dst, GLC_CN_SURFACE_DRAWABLE_CURRENT)) {
    glc_surface_pop_current (dst);
    return;
  }

  glc_set_operator (op);

  glShadeModel (GL_FLAT);

  for (index = 0; index < n_color_traps; index++) {
    
    if (!COLORTRAPEZOID_VALID (&color_traps[index]))
      continue;
    
    if (memcmp (&color_traps[index].top.right_color,
                &color_traps[index].top.left_color, sizeof (glc_color_t)) ||
        memcmp (&color_traps[index].bottom.right_color,
                &color_traps[index].bottom.left_color, sizeof (glc_color_t)) ||
        memcmp (&color_traps[index].top.left_color,
                &color_traps[index].bottom.left_color, sizeof (glc_color_t))) {
      glShadeModel (GL_SMOOTH);
      shade = 1;
      break;
    }
  }
  
  glBegin (GL_QUADS);
  
  for (; n_color_traps; n_color_traps--, color_traps++) {
    
    if (!COLORTRAPEZOID_VALID (color_traps))
      continue;

    glColor4us (color_traps->bottom.left_color.red,
                color_traps->bottom.left_color.green,
                color_traps->bottom.left_color.blue,
                color_traps->bottom.left_color.alpha);
    glVertex2d (FIXED_TO_DOUBLE (color_traps->bottom.left),
                FIXED_TO_DOUBLE (color_traps->bottom.y));

    if (shade)
      glColor4us (color_traps->top.left_color.red,
                  color_traps->top.left_color.green,
                  color_traps->top.left_color.blue,
                  color_traps->top.left_color.alpha);
    glVertex2d (FIXED_TO_DOUBLE (color_traps->top.left),
                FIXED_TO_DOUBLE (color_traps->top.y));

    if (shade)
      glColor4us (color_traps->top.right_color.red,
                  color_traps->top.right_color.green,
                  color_traps->top.right_color.blue,
                  color_traps->top.right_color.alpha);
    glVertex2d (FIXED_TO_DOUBLE (color_traps->top.right),
                FIXED_TO_DOUBLE (color_traps->top.y));

    if (shade)
      glColor4us (color_traps->bottom.right_color.red,
                  color_traps->bottom.right_color.green,
                  color_traps->bottom.right_color.blue,
                  color_traps->bottom.right_color.alpha);
    glVertex2d (FIXED_TO_DOUBLE (color_traps->bottom.right),
                FIXED_TO_DOUBLE (color_traps->bottom.y));
  }
  
  glEnd ();
  
  glShadeModel (GL_FLAT);

  glc_surface_dirty (dst, &bounds);
  glc_surface_pop_current (dst);
}
slim_hidden_def(glc_color_trapezoids);
