/*
 * Copyright  2004 David Reveman, Peter Nilsson
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the names of
 * David Reveman and Peter Nilsson not be used in advertising or
 * publicity pertaining to distribution of the software without
 * specific, written prior permission. David Reveman and Peter Nilsson
 * makes no representations about the suitability of this software for
 * any purpose. It is provided "as is" without express or implied warranty.
 *
 * DAVID REVEMAN AND PETER NILSSON DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL DAVID REVEMAN AND
 * PETER NILSSON BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: David Reveman <c99drn@cs.umu.se>
 *          Peter Nilsson <c99pnn@cs.umu.se>
 */

#ifdef HAVE_CONFIG_H
#  include "../config.h"
#endif

#include "glcint.h"

static void
glc_point_fixed_bounds (int n_point,
                        const glc_point_fixed_t *points,
                        glc_region_box_t *box)
{
  box->x1 = FIXED_TO_INT (points->x);
  box->x2 = FIXED_TO_INT (FIXED_CEIL (points->x));
  box->y1 = FIXED_TO_INT (points->y);
  box->y2 = FIXED_TO_INT (FIXED_CEIL (points->y));
  points++;
  n_point--;
    
  while (n_point-- > 0) {
    int x1 = FIXED_TO_INT (points->x);
    int x2 = FIXED_TO_INT (FIXED_CEIL (points->x));
    int y1 = FIXED_TO_INT (points->y);
    int y2 = FIXED_TO_INT (FIXED_CEIL (points->y));
      
    if (x1 < box->x1)
      box->x1 = x1;
    else if (x2 > box->x2)
      box->x2 = x2;
    if (y1 < box->y1)
      box->y1 = y1;
    else if (y2 > box->y2)
      box->y2 = y2;
    points++;
  }
}

void
glc_composite_triangles (glc_operator_t op,
                         glc_surface_t *src,
                         glc_surface_t *dst,
                         int x_src,
                         int y_src,
                         const glc_triangle_t *tris,
                         int n_tris)
{
  glc_surface_t *mask;
  glc_region_box_t tri_bounds, src_bounds, dst_bounds, bounds;
  double x_draw_offset, y_draw_offset;
  int x_dst, y_dst;

  x_dst = tris[0].p1.x >> 16;
  y_dst = tris[0].p1.y >> 16;

  glc_point_fixed_bounds (n_tris * 3, (glc_point_fixed_t *) tris, &tri_bounds);
  glc_surface_bounds (src, &src_bounds);
  glc_surface_bounds (dst, &dst_bounds);

  src_bounds.x1 += (x_dst - x_src);
  src_bounds.y1 += (y_dst - y_src);
  src_bounds.x2 += (x_dst - x_src);
  src_bounds.y2 += (y_dst - y_src);

  glc_intersect_region (&src_bounds, &tri_bounds, &bounds);
  glc_intersect_region (&dst_bounds, &bounds, &bounds);

  if ((bounds.x2 - bounds.x1) <= 0 || (bounds.y2 - bounds.y1) <= 0)
    return;
  
  x_draw_offset = (double) -bounds.x1;
  y_draw_offset = (double) -bounds.y1;
  
  mask = glc_surface_create_similar (dst, GLC_STANDARD_A8,
                                     bounds.x2 - bounds.x1,
                                     bounds.y2 - bounds.y1);
  
  if (!mask) {
    glc_surface_status_add (dst, GLC_STATUS_NOT_SUPPORTED_MASK);
    return;
  }

  mask->implicit_mask = 1;

  if (!glc_surface_push_current (mask, GLC_CN_SURFACE_DRAWABLE_CURRENT)) {
    glc_surface_pop_current (mask);
    return;
  }

  glClearColor (0.0, 0.0, 0.0, 0.0);
  glClear (GL_COLOR_BUFFER_BIT);
  glColor4f (1.0, 1.0, 1.0, 1.0);

  glc_set_operator (GLC_OPERATOR_SRC);

  if (mask->polyedge == GLC_POLYEDGE_SMOOTH) {
    if (!mask->anti_aliasing) {
      glEnable (GL_POLYGON_SMOOTH);
      glEnable (GL_BLEND);
      glHint(GL_POLYGON_SMOOTH, GL_NICEST);
      glBlendFunc (GL_SRC_ALPHA_SATURATE, GL_ONE);
    }
  }
  
  glBegin (GL_TRIANGLES);
  
  for (; n_tris; n_tris--, tris++) {
    glVertex2d (x_draw_offset + FIXED_TO_DOUBLE (tris->p1.x),
                y_draw_offset + FIXED_TO_DOUBLE (tris->p1.y));
    glVertex2d (x_draw_offset + FIXED_TO_DOUBLE (tris->p2.x),
                y_draw_offset + FIXED_TO_DOUBLE (tris->p2.y));
    glVertex2d (x_draw_offset + FIXED_TO_DOUBLE (tris->p3.x),
                y_draw_offset + FIXED_TO_DOUBLE (tris->p3.y));
  }
  
  glEnd ();

  if (mask->polyedge == GLC_POLYEDGE_SMOOTH) {
    if (!mask->anti_aliasing) {
      glDisable (GL_POLYGON_SMOOTH);
      glDisable (GL_BLEND);
    }
  }

  glc_surface_dirty (mask, NULL);
  glc_surface_pop_current (mask);

  glc_composite (op,
                 src,
                 mask,
                 dst,
                 x_src + tri_bounds.x1 - x_dst,
                 y_src + tri_bounds.y1 - y_dst,
                 0, 0,
                 bounds.x1, bounds.y1,
                 mask->width, mask->height);
  
  glc_surface_destroy (mask);
}
slim_hidden_def(glc_composite_triangles);
     
void
glc_composite_tri_strip (glc_operator_t op,
                         glc_surface_t *src,
                         glc_surface_t *dst,
                         int x_src,
                         int y_src,
                         const glc_point_fixed_t *points,
                         int n_points)
{
  glc_surface_t *mask;
  glc_region_box_t tri_bounds, src_bounds, dst_bounds, bounds;
  double x_draw_offset, y_draw_offset;
  int x_dst, y_dst;

  if (n_points < 3)
    return;

  x_dst = points[0].x >> 16;
  y_dst = points[0].y >> 16;

  glc_point_fixed_bounds (n_points, points, &tri_bounds);
  glc_surface_bounds (src, &src_bounds);
  glc_surface_bounds (dst, &dst_bounds);

  src_bounds.x1 += (x_dst - x_src);
  src_bounds.y1 += (y_dst - y_src);
  src_bounds.x2 += (x_dst - x_src);
  src_bounds.y2 += (y_dst - y_src);

  glc_intersect_region (&src_bounds, &tri_bounds, &bounds);
  glc_intersect_region (&dst_bounds, &bounds, &bounds);

  if ((bounds.x2 - bounds.x1) <= 0 || (bounds.y2 - bounds.y1) <= 0)
    return;
  
  x_draw_offset = (double) -bounds.x1;
  y_draw_offset = (double) -bounds.y1;
  
  mask = glc_surface_create_similar (dst, GLC_STANDARD_A8,
                                     bounds.x2 - bounds.x1,
                                     bounds.y2 - bounds.y1);
  
  if (!mask) {
    glc_surface_status_add (dst, GLC_STATUS_NOT_SUPPORTED_MASK);
    return;
  }

  mask->implicit_mask = 1;

  if (!glc_surface_push_current (mask, GLC_CN_SURFACE_DRAWABLE_CURRENT)) {
    glc_surface_pop_current (mask);
    return;
  }

  glClearColor (0.0, 0.0, 0.0, 0.0);
  glClear (GL_COLOR_BUFFER_BIT);
  glColor4f (1.0, 1.0, 1.0, 1.0);

  glc_set_operator (GLC_OPERATOR_SRC);

  if (mask->polyedge == GLC_POLYEDGE_SMOOTH) {
    if (!mask->anti_aliasing) {
      glEnable (GL_POLYGON_SMOOTH);
      glEnable (GL_BLEND);
      glHint(GL_POLYGON_SMOOTH, GL_NICEST);
      glBlendFunc (GL_SRC_ALPHA_SATURATE, GL_ONE);
    }
  }

  glBegin (GL_TRIANGLE_STRIP);
  
  for (; n_points; n_points--, points++)
    glVertex2d (x_draw_offset + FIXED_TO_DOUBLE (points->x),
                y_draw_offset + FIXED_TO_DOUBLE (points->y));
  
  glEnd ();

  if (mask->polyedge == GLC_POLYEDGE_SMOOTH) {
    if (!mask->anti_aliasing) {
      glDisable (GL_POLYGON_SMOOTH);
      glDisable (GL_BLEND);
    }
  }

  glc_surface_dirty (mask, NULL);
  glc_surface_pop_current (mask);
  
  glc_composite (op,
                 src,
                 mask,
                 dst,
                 x_src + tri_bounds.x1 - x_dst,
                 y_src + tri_bounds.y1 - y_dst,
                 0, 0,
                 bounds.x1, bounds.y1,
                 mask->width, mask->height);

  glc_surface_destroy (mask);
}
slim_hidden_def(glc_composite_tri_strip);

void
glc_composite_tri_fan (glc_operator_t op,
                       glc_surface_t *src,
                       glc_surface_t *dst,
                       int x_src,
                       int y_src,
                       const glc_point_fixed_t *points,
                       int n_points)
{
  glc_surface_t *mask;
  glc_region_box_t tri_bounds, src_bounds, dst_bounds, bounds;
  double x_draw_offset, y_draw_offset;
  int x_dst, y_dst;

  if (n_points < 3)
    return;

  x_dst = points[0].x >> 16;
  y_dst = points[0].y >> 16;

  glc_point_fixed_bounds (n_points, points, &tri_bounds);
  glc_surface_bounds (src, &src_bounds);
  glc_surface_bounds (dst, &dst_bounds);

  src_bounds.x1 += (x_dst - x_src);
  src_bounds.y1 += (y_dst - y_src);
  src_bounds.x2 += (x_dst - x_src);
  src_bounds.y2 += (y_dst - y_src);

  glc_intersect_region (&src_bounds, &tri_bounds, &bounds);
  glc_intersect_region (&dst_bounds, &bounds, &bounds);

  if ((bounds.x2 - bounds.x1) <= 0 || (bounds.y2 - bounds.y1) <= 0)
    return;
  
  x_draw_offset = (double) -bounds.x1;
  y_draw_offset = (double) -bounds.y1;
  
  mask = glc_surface_create_similar (dst, GLC_STANDARD_A8,
                                     bounds.x2 - bounds.x1,
                                     bounds.y2 - bounds.y1);
  
  if (!mask) {
    glc_surface_status_add (dst, GLC_STATUS_NOT_SUPPORTED_MASK);
    return;
  }
  
  mask->implicit_mask = 1;

  if (!glc_surface_push_current (mask, GLC_CN_SURFACE_DRAWABLE_CURRENT)) {
    glc_surface_pop_current (mask);
    return;
  }

  glClearColor (0.0, 0.0, 0.0, 0.0);
  glClear (GL_COLOR_BUFFER_BIT);
  glColor4f (1.0, 1.0, 1.0, 1.0);

  glc_set_operator (GLC_OPERATOR_SRC);

  if (mask->polyedge == GLC_POLYEDGE_SMOOTH) {
    if (!mask->anti_aliasing) {
      glEnable (GL_POLYGON_SMOOTH);
      glEnable (GL_BLEND);
      glHint(GL_POLYGON_SMOOTH, GL_NICEST);
      glBlendFunc (GL_SRC_ALPHA_SATURATE, GL_ONE);
    }
  }
  
  glBegin (GL_TRIANGLE_FAN);

  for (; n_points; n_points--, points++)
    glVertex2d (x_draw_offset + FIXED_TO_DOUBLE (points->x),
                y_draw_offset + FIXED_TO_DOUBLE (points->y));
  
  glEnd ();

  if (mask->polyedge == GLC_POLYEDGE_SMOOTH) {
    if (!mask->anti_aliasing) {
      glDisable (GL_POLYGON_SMOOTH);
      glDisable (GL_BLEND);
    }
  }

  glc_surface_dirty (mask, NULL);
  glc_surface_pop_current (mask);

  glc_composite (op,
                 src,
                 mask,
                 dst,
                 x_src + tri_bounds.x1 - x_dst,
                 y_src + tri_bounds.y1 - y_dst,
                 0, 0,
                 bounds.x1, bounds.y1,
                 mask->width, mask->height);

  glc_surface_destroy (mask);
}
slim_hidden_def(glc_composite_tri_fan);

static void
glc_color_triangle_bounds (int n_color_tris,
                           const glc_color_triangle_t *color_tris,
                           glc_region_box_t *box)
{
  box->x1 = MIN (MIN (FIXED_TO_INT (color_tris->p1.point.x),
                      FIXED_TO_INT (color_tris->p2.point.x)),
                 FIXED_TO_INT (color_tris->p3.point.x));

  box->x2 = MAX (MAX (FIXED_TO_INT (FIXED_CEIL (color_tris->p1.point.x)),
                      FIXED_TO_INT (FIXED_CEIL (color_tris->p2.point.x))),
                 FIXED_TO_INT (FIXED_CEIL (color_tris->p3.point.x)));
  
  box->y1 = MIN (MIN (FIXED_TO_INT (color_tris->p1.point.y),
                      FIXED_TO_INT (color_tris->p2.point.y)),
                 FIXED_TO_INT (color_tris->p3.point.y));
  
  box->y2 = MAX (MAX (FIXED_TO_INT (FIXED_CEIL (color_tris->p1.point.y)),
                      FIXED_TO_INT (FIXED_CEIL (color_tris->p2.point.y))),
                 FIXED_TO_INT (FIXED_CEIL (color_tris->p3.point.y)));
  
  color_tris++;
  n_color_tris--;

  while (n_color_tris-- > 0) {
    int16_t x1, y1, x2, y2;

    x1 = MIN (MIN (FIXED_TO_INT (color_tris->p1.point.x),
                   FIXED_TO_INT (color_tris->p2.point.x)),
              FIXED_TO_INT (color_tris->p3.point.x));
    if (x1 < box->x1)
      box->x1 = x1;
    
    x2 = MAX (MAX (FIXED_TO_INT (FIXED_CEIL (color_tris->p1.point.x)),
                   FIXED_TO_INT (FIXED_CEIL (color_tris->p2.point.x))),
              FIXED_TO_INT (FIXED_CEIL (color_tris->p3.point.x)));
    if (x2 > box->x2)
      box->x2 = x2;
    
    y1 = MIN (MIN (FIXED_TO_INT (color_tris->p1.point.y),
                   FIXED_TO_INT (color_tris->p2.point.y)),
              FIXED_TO_INT (color_tris->p3.point.y));
    if (y1 < box->y1)
      box->y1 = y1;
    
    y2 = MAX (MAX (FIXED_TO_INT (FIXED_CEIL (color_tris->p1.point.y)),
                   FIXED_TO_INT (FIXED_CEIL (color_tris->p2.point.y))),
              FIXED_TO_INT (FIXED_CEIL (color_tris->p3.point.y)));
    if (y2 > box->y2)
      box->y2 = y2;
    
    color_tris++;
  }
}


void
glc_color_triangles (glc_operator_t op,
                     glc_surface_t *dst,
                     const glc_color_triangle_t *color_tris,
                     int n_color_tris)
{
  glc_region_box_t bounds;
  int index;
  glc_bool_t shade = 0;

  glc_color_triangle_bounds (n_color_tris, color_tris, &bounds);
  if (bounds.y1 >= bounds.y2 || bounds.x1 >= bounds.x2 ||
      bounds.x1 > dst->width || bounds.y1 > dst->height ||
      bounds.x2 < 0 || bounds.y2 < 0)
    return;
  
  if (!glc_surface_push_current (dst, GLC_CN_SURFACE_DRAWABLE_CURRENT)) {
    glc_surface_pop_current (dst);
    return;
  }

  glc_set_operator (op);

  glShadeModel (GL_FLAT);
  for (index = 0; index < n_color_tris; index++) {
    if (memcmp (&color_tris[index].p1.color,
                &color_tris[index].p2.color, sizeof (glc_color_t)) ||
        memcmp (&color_tris[index].p2.color,
                &color_tris[index].p3.color, sizeof (glc_color_t))) {
      glShadeModel (GL_SMOOTH);
      shade = 1;
      break;
    }
  }

  glBegin (GL_TRIANGLES);

  for (; n_color_tris; n_color_tris--, color_tris++) {    
    glColor4us (color_tris->p1.color.red,
                color_tris->p1.color.green,
                color_tris->p1.color.blue,
                color_tris->p1.color.alpha);
    glVertex2d (FIXED_TO_DOUBLE (color_tris->p1.point.x),
                FIXED_TO_DOUBLE (color_tris->p1.point.y));
    
    if (shade)
      glColor4us (color_tris->p2.color.red,
                  color_tris->p2.color.green,
                  color_tris->p2.color.blue,
                  color_tris->p2.color.alpha);
    glVertex2d (FIXED_TO_DOUBLE (color_tris->p2.point.x),
                FIXED_TO_DOUBLE (color_tris->p2.point.y));

    if (shade)
      glColor4us (color_tris->p3.color.red,
                  color_tris->p3.color.green,
                  color_tris->p3.color.blue,
                  color_tris->p3.color.alpha);
    glVertex2d (FIXED_TO_DOUBLE (color_tris->p3.point.x),
                FIXED_TO_DOUBLE (color_tris->p3.point.y));
  }

  glEnd ();

  glShadeModel (GL_FLAT);

  glc_surface_dirty (dst, &bounds);
  glc_surface_pop_current (dst);
}
slim_hidden_def(glc_color_triangles);
